import numpy as np
import pywt
import scipy.stats


def waveletDenoise(data):
    # data is num_neurons x time_frames

    wavelet = pywt.Wavelet('db4')

    # Determine the maximum number of possible levels for image
    dlen = wavelet.dec_len
    wavelet_levels = pywt.dwt_max_level(data.shape[1], wavelet)

    # Skip coarsest wavelet scales (see Notes in docstring).
    wavelet_levels = max(wavelet_levels - 3, 1)

    data_denoise = np.zeros(np.shape(data))

    shift = 4
    for c in np.arange(-shift, shift + 1):
        data_shift = np.roll(data, c, 1)
        for i in range(np.shape(data)[0]):
            coeffs = pywt.wavedecn(data_shift[i, :], wavelet=wavelet,
                                   level=wavelet_levels)
            # Detail coefficients at each decomposition level
            dcoeffs = coeffs[1:]
            detail_coeffs = dcoeffs[-1]['d']
            # rescaling using a single estimation of level noise based on first level coefficients.
            # Consider regions with detail coefficients exactly zero to be masked out
            # detail_coeffs = detail_coeffs[np.nonzero(detail_coeffs)]
            # 75th quantile of the underlying, symmetric noise distribution
            denom = scipy.stats.norm.ppf(0.75)
            sigma = np.median(np.abs(detail_coeffs)) / denom
            np.shape(sigma)
            sigma_mat = np.tile(sigma, (wavelet_levels, 1))
            np.shape(sigma_mat)

            tot_num_coeffs = pywt.wavedecn_size(coeffs)
            # universal threshold
            threshold = np.sqrt(2 * np.log(tot_num_coeffs))
            threshold = sigma * threshold

            denoised_detail = [{key: pywt.threshold(level[key],
                                                    value=threshold,
                                                    mode='hard') for key in level}
                               for level in dcoeffs]

            # Dict of unique threshold coefficients for each detail coeff. array

            denoised_coeffs = [coeffs[0]] + denoised_detail

            data_denoise[i, :] = data_denoise[i, :] + np.roll(
                pywt.waverecn(denoised_coeffs, wavelet), -c)[:data_denoise.shape[1]]

    data_denoise = data_denoise / (2 * shift + 1)
    return data_denoise



from CIDAN.GUI.Data_Interaction.Signals import *


class Thread(QThread):
    def __init__(self, main_widget, parent=None):
        QThread.__init__(self, parent)
        self.exiting = False
        self.main_widget = main_widget
        self.signal = BoolSignal()

    @property
    def data_handler(self):
        return self.main_widget.data_handler
    def run(self):
        pass

    def endThread(self, success):
        pass

    # def __del__(self):
    #     self.exiting = True
    #     self.wait()



import argparse
import logging
import sys

from qtpy.QtWidgets import QApplication

from CIDAN.GUI import MainWindow
from CIDAN.GUI.Data_Interaction import DataHandler


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--dev", type=bool, default=False, required=False,
                        help="Enable Developement mode")
    LOG_FILENAME = 'log.out'
    parser.add_argument("-lp", "--logpath", type=str, default=LOG_FILENAME,
                        required=False, help="Path to save log file")
    parser.add_argument("--headless", type=bool, default=False,
                        required=False, help="Whether to run software headless")
    parser.add_argument("-p", "--parameter", type=str, default=False,
                        required=False, help="Path to parameter file")
    parser.add_argument("-ll", "--loglevel", type=str, default="error",
                        required=False,
                        choices=['debug', 'info', 'warning', 'error', 'critical'])
    args = parser.parse_args()
    LEVELS = {'debug': logging.DEBUG,
              'info': logging.INFO,
              'warning': logging.WARNING,
              'error': logging.ERROR,
             'critical': logging.CRITICAL}

    level = LEVELS.get(args.loglevel, logging.NOTSET)
    logging.basicConfig(filename=args.logpath, level=level)
    logger = logging.getLogger("CIDAN")
    logger.debug("Program started")
    if args.headless:
        data_handler = DataHandler.DataHandler("",
                                               "",
                                               trials=[],
                                               save_dir_already_created=True,
                                               parameter_file=args.parameter,
                                               load_into_mem=True)
        data_handler.calculate_filters()
        data_handler.calculate_roi_extraction()
        data_handler.export()
    else:
        app = QApplication([])
        app.setApplicationName("CIDAN")
        widget = MainWindow.MainWindow(dev=args.dev)

        sys.exit(app.exec_())


if __name__ == '__main__':
    main()



from qtpy.QtWidgets import *

from CIDAN.GUI.Inputs.BoolInput import BoolInput
from CIDAN.GUI.Inputs.FloatInput import FloatInput
from CIDAN.GUI.Inputs.IntInput import IntInput
from CIDAN.GUI.Inputs.IntRangeInput import IntRangeInput


class SettingBlockModule(QFrame):
    """
    A tab of settings in SettingsModule

    These are all the tabs in each of the settings area in the GUI, just specify each
    input in a list and it will add them together
    """
    def __init__(self, name, input_list):
        """
        Initializes the list
        Parameters
        ----------
        name : str
            Name of the list of modules
        input_list : List[Input]
            The list of inputs that are part of this section of settings
        """
        super().__init__()
        self.name = name
        self.input_list = input_list
        self.layout = QVBoxLayout()
        # Type options are int, float, bool, and str
        for input in input_list:
            self.layout.addWidget(input)
        # self.setMaximumWidth(300)
        self.setLayout(self.layout)


def filter_setting_block(main_widget):
    data_handler = main_widget.data_handler
    return SettingBlockModule("Filter Settings",
                              [BoolInput(display_name="Local Spatial Denoising:",
                                         program_name="localSpatialDenoising",
                                         on_change_function=lambda x,
                                                                   y: data_handler.change_filter_param(
                                             x, y),
                                         default_val=data_handler.filter_params[
                                             "localSpatialDenoising"],
                                         tool_tip="Local Spatial Denoising helps smooth out the noise in the data by averaging ",
                                         ),

                               BoolInput(
                                   display_name="Z-score:",
                                   program_name="z_score",
                                   on_change_function=lambda x,
                                                             y: data_handler.change_filter_param(
                                       x, y),
                                   default_val=
                                   data_handler.filter_params["z_score"],
                                   tool_tip="Applies a z-score for each pixel across each of the timesteps",
                               ),
                               BoolInput(display_name="Histogram Equalization Method:",
                                         program_name="hist_eq",
                                         on_change_function=lambda x,
                                                                   y: data_handler.change_filter_param(
                                             x, y),
                                         default_val=data_handler.filter_params[
                                             "hist_eq"],
                                         tool_tip="Applies a histogram equalization method(a more advanced z score",
                                         ),
                               BoolInput(display_name="Median filter:",
                                         program_name="median_filter",
                                         on_change_function=lambda x,
                                                                   y: data_handler.change_filter_param(
                                             x, y),
                                         default_val=data_handler.filter_params[
                                             "median_filter"],
                                         tool_tip="Applies a 3D median filter to each timestep. This helps denoise the image. ",
                                         ),
                               IntInput(
                                   display_name="Median filter size:",
                                   program_name="median_filter_size",
                                   on_change_function=lambda x,
                                                             y: data_handler.change_filter_param(
                                       x, y),
                                   default_val=
                                   data_handler.filter_params[
                                       "median_filter_size"],
                                   tool_tip="The size of the 3D median filter. Recomended size: 3",
                                   min=1, max=50, step=1),
                               BoolInput(display_name="PCA:",
                                         program_name="pca",
                                         on_change_function=lambda x,
                                                                   y: data_handler.change_filter_param(
                                             x, y),
                                         default_val=data_handler.filter_params[
                                             "pca"],
                                         tool_tip="Applies PCA decomposition to the dataset (PCA runs after other filters). \n Both speeds along other calculations and helps remove background noise. \n Time traces are still calculated on filtered stack not PCA stack.",
                                         ),
                               # FloatInput(
                               #     display_name="PCA expression threshold",
                               #     program_name="pca_threshold",
                               #     on_change_function=lambda x,
                               #                               y: data_handler.change_filter_param(
                               #         x, y),
                               #     default_val=
                               #     data_handler.filter_params[
                               #         "pca_threshold"],
                               #     tool_tip="The percentage of the variance that the PCA will express",
                               #     min=0.001, max=.999, step=.001),
                               ])


def dataset_setting_block(main_widget):
    data_handler = main_widget.data_handler
    return SettingBlockModule("Advanced Settings",
                              [BoolInput(display_name="Slice stack:",
                                         program_name="slice_stack",
                                         on_change_function=lambda x,
                                                                   y: data_handler.change_dataset_param(
                                             x, y),
                                         default_val=data_handler.dataset_params[
                                             "slice_stack"],
                                         tool_tip="Used to Select only "
                                                  "every x timestep",
                                         display_tool_tip=False),
                               IntInput(display_name="Slice every:",
                                        program_name="slice_every",
                                        on_change_function=lambda x,
                                                                  y: data_handler.change_dataset_param(
                                            x, y),
                                        default_val=data_handler.dataset_params[
                                            "slice_every"],
                                        tool_tip="Select every x timesteps",
                                        min=1, max=100, step=1),
                               IntInput(display_name="Slice start:",
                                        program_name="slice_start",
                                        on_change_function=lambda x,
                                                                  y: data_handler.change_dataset_param(
                                            x, y),
                                        tool_tip="Start selecting on x timestep",
                                        default_val=data_handler.dataset_params[
                                            "slice_start"],
                                        min=0, max=10000, step=1)
                               ])


def dataset_setting_block_crop(main_widget):
    data_handler = main_widget.data_handler
    return SettingBlockModule("Crop Settings",
                              [BoolInput(display_name="Crop stack:",
                                         program_name="crop_stack",
                                         on_change_function=lambda x,
                                                                   y: data_handler.change_dataset_param(
                                             x, y),
                                         default_val=data_handler.dataset_params[
                                             "crop_stack"],
                                         tool_tip="Used to crop image stack",
                                         display_tool_tip=False),
                               IntRangeInput(display_name="Crop Row:",
                                             program_name="crop_y",
                                             on_change_function=lambda x,
                                                                       y: data_handler.change_dataset_param(
                                                 x, y),
                                             default_val=data_handler.dataset_params[
                                                 "crop_y"],
                                             tool_tip="Crop rows",
                                             min=0, max=10000, step=1),
                               IntRangeInput(display_name="Crop Column:",
                                             # different because the way we display images is weird x is first dim y is second dim
                                             program_name="crop_x",
                                             on_change_function=lambda x,
                                                                       y: data_handler.change_dataset_param(
                                                 x, y),
                                             default_val=data_handler.dataset_params[
                                                 "crop_x"],
                                             tool_tip="Crop columns",
                                             min=0, max=100000, step=1),
                               ] + ([BoolInput(
                                  display_name="Split into Time Blocks(recomended):",
                                  program_name="trial_split",
                                  on_change_function=lambda x,
                                                            y: data_handler.change_dataset_param(
                                      x, y),
                                  default_val=data_handler.dataset_params[
                                      "trial_split"],
                                  tool_tip="Splits the timesteps into separate time blocks better for processing "
                                           "every x timestep. This setting is recomended for all large datasets",
                                  display_tool_tip=False),
                                                                      IntInput(
                                                                          display_name="Time Block Length",
                                                                          program_name="trial_length",
                                                                          on_change_function=lambda x,
                                                                             y: data_handler.change_dataset_param(
                                                       x, y),
                                                                          default_val=
                                                   data_handler.dataset_params[
                                                       "trial_length"],
                                                                          tool_tip="Length of each timeblock",
                                                                          display_tool_tip=False,
                                                                          min=25,
                                                                          max=2000,
                                                                          step=1)] if
                                                                  data_handler.dataset_params[
                                                                      "original_folder_trial_split"] != "" else []))


def multiprocessing_settings_block(main_widget):
    data_handler = main_widget.data_handler
    return SettingBlockModule("Multiprocessing Settings", [

        IntInput(
            display_name="Number of spatial boxes:",
            program_name="total_num_spatial_boxes",
            on_change_function=lambda x, y: data_handler.change_box_param(
                x, y),
            default_val=
            data_handler.box_params[
                "total_num_spatial_boxes"],
            tool_tip="Used to break the calculation into smaller overlapping parts.\n "
                     "This number must be square. The number of spatial boxes per side will be equal to square of this number.",
            min=1, max=10000, step=1)
        ,
        IntInput(
            display_name="Spatial overlap:",
            program_name="spatial_overlap",
            on_change_function=lambda x, y: data_handler.change_box_param(x, y),
            default_val=
            data_handler.box_params[
                "spatial_overlap"],
            tool_tip="Number of pixels to overlap each box.",
            min=1, max=10000, step=1)])


def roi_extraction_settings_block(main_widget):
    # TODO fix input into spatial box number
    data_handler = main_widget.data_handler
    return SettingBlockModule("ROI Extraction Settings",
                              [
                                  IntInput(
                                      display_name="Max Number of ROIs per spatial box:",
                                      program_name="num_rois",
                                      on_change_function=lambda x,
                                                                y: data_handler.change_roi_extraction_param(
                                          x, y),
                                      default_val=data_handler.roi_extraction_params[
                                          "num_rois"],
                                      tool_tip="Max number of ROIs to select for each spatial box, increase this number if not detecting all ROIs",
                                      min=0, max=10000, step=1),
                                  IntInput(
                                      display_name="ROI size minimum:",
                                      program_name="roi_size_min",
                                      on_change_function=lambda x,
                                                                y: data_handler.change_roi_extraction_param(
                                          x, y),
                                      default_val=
                                      data_handler.roi_extraction_params[
                                          "roi_size_min"],
                                      tool_tip="Minimum size in pixels for a region of interest",
                                      min=1, max=10000, step=1),
                                  IntInput(
                                      display_name="ROI size maximum:",
                                      program_name="roi_size_max",
                                      on_change_function=lambda x,
                                                                y: data_handler.change_roi_extraction_param(
                                          x, y),
                                      default_val=
                                      data_handler.roi_extraction_params[
                                          "roi_size_max"],
                                      tool_tip="Maximum size in pixels for a region of interest",
                                      min=1, max=1000000, step=1),
                                  # BoolInput(
                                  #     display_name="Run refinement step:",
                                  #     program_name="refinement",
                                  #     on_change_function=lambda x,
                                  #                               y: data_handler.change_roi_extraction_param(
                                  #         x, y),
                                  #     default_val=
                                  #     data_handler.roi_extraction_params[
                                  #         "refinement"],
                                  #     tool_tip="Whether to run the refinement step, greatly improves results"),
                                  FloatInput(
                                      display_name="Merge temporal coefficient:",
                                      program_name="merge_temporal_coef",
                                      on_change_function=lambda x,
                                                                y: data_handler.change_roi_extraction_param(
                                          x, y),
                                      default_val=
                                      data_handler.roi_extraction_params[
                                          "merge_temporal_coef"],
                                      tool_tip="The coefficient that determines if two overlapping regions are merged based on their temporal correlation",
                                      min=0, max=1, step=.01)
                                  #     BoolInput(
                                  #         display_name="Fill holes:",
                                  #         program_name="fill_holes",
                                  #         on_change_function=lambda x,
                                  #                                   y: data_handler.change_roi_extraction_param(
                                  #             x, y),
                                  #         default_val=
                                  #         data_handler.roi_extraction_params[
                                  #             "fill_holes"],
                                  #         tool_tip="Whether to fill holes in each roi"),
                                  #
                              ]
                              )


def roi_advanced_settings_block(main_widget):
    # TODO fix input into spatial box number
    data_handler = main_widget.data_handler
    return SettingBlockModule("Advanced Settings",
                              [IntInput(display_name="Number of eigen vectors:",
                                        program_name="num_eig",
                                        on_change_function=lambda x,
                                                                  y: data_handler.change_eigen_param(
                                            x, y),
                                        default_val=data_handler.eigen_params[
                                            "num_eig"],
                                        tool_tip="Number of eigen vectors to generate. Increase if eigen norm image isn't showing all rois",
                                        min=1, max=10000, step=1),
                               FloatInput(display_name="Eigen threshold value:",
                                          program_name="eigen_threshold_value",
                                          on_change_function=lambda x,
                                                                    y: data_handler.change_roi_extraction_param(
                                              x, y),
                                          default_val=
                                          data_handler.roi_extraction_params[
                                              "eigen_threshold_value"],
                                          tool_tip="Number of eigen vectors to select for each ROI. Multiply number of trials by the number of eigen vectors \n if this number is below 300, then it might help to increase this value to .3 or .5",
                                          min=0, max=1, step=.01),
                               # FloatInput(display_name="Elbow threshold value:",
                               #            program_name="elbow_threshold_value",
                               #            on_change_function=lambda x,
                               #                                      y: data_handler.change_roi_extraction_param(
                               #                x, y),
                               #            default_val=
                               #            data_handler.roi_extraction_params[
                               #                "elbow_threshold_value"],
                               #            tool_tip="Number of eigen vectors to select at each point",
                               #            min=0, max=1.5, step=.01),
                               # BoolInput(
                               #     display_name="Merge ROIs:",
                               #     program_name="merge",
                               #     on_change_function=lambda x,
                               #                               y: data_handler.change_roi_extraction_param(
                               #         x, y),
                               #     default_val=
                               #     data_handler.roi_extraction_params[
                               #         "merge"],
                               #     tool_tip="Whether to merge rois with similar time traces"),
                               IntInput(display_name="ROI Eccentricity Threshold:",
                                        program_name="roi_circ_threshold",
                                        on_change_function=lambda x,
                                                                    y: data_handler.change_roi_extraction_param(
                                              x, y),
                                        default_val=
                                          data_handler.roi_extraction_params[
                                              "roi_circ_threshold"],
                                        tool_tip="Thresholds the rois based on how circular they are. Lowering the number would allow more ROIs through",
                                        min=0, max=100, step=1),
                               # IntInput(display_name="Number of connections:",
                               #          program_name="connections",
                               #          on_change_function=lambda x,
                               #                                    y: data_handler.change_eigen_param(
                               #              x, y),
                               #          default_val=data_handler.eigen_params[
                               #              "connections"],
                               #          tool_tip="Number of eigen vectors to generate",
                               #          min=1, max=10000, step=1),
                               # IntInput(display_name="accuracy:",
                               #          program_name="accuracy",
                               #          on_change_function=lambda x,
                               #                                    y: data_handler.change_eigen_param(
                               #              x, y),
                               #          default_val=data_handler.eigen_params[
                               #              "accuracy"],
                               #          tool_tip="Number of eigen vectors to generate",
                               #          min=1, max=10000, step=1),
                               # IntInput(display_name="Number of knn:",
                               #          program_name="knn",
                               #          on_change_function=lambda x,
                               #                                    y: data_handler.change_eigen_param(
                               #              x, y),
                               #          default_val=data_handler.eigen_params[
                               #              "knn"],
                               #          tool_tip="Number of eigen vectors to generate",
                               #          min=1, max=10000, step=1),
                               IntInput(display_name="Eigen Accuracy:",
                                        program_name="eigen_accuracy",
                                        on_change_function=lambda x,
                                                                  y: data_handler.change_eigen_param(
                                            x, y),
                                        default_val=data_handler.eigen_params[
                                            "eigen_accuracy"],
                                        tool_tip="10^-x accuracy for each eigen vector, If there are lines in the eigen norm image increase this to 7 or 8",
                                        min=1, max=10000, step=1),
                               IntInput(display_name="Number of iterations:",
                                        program_name="max_iter",
                                        on_change_function=lambda x,
                                                                  y: data_handler.change_roi_extraction_param(
                                            x, y),
                                        default_val=data_handler.roi_extraction_params[
                                            "max_iter"],
                                        tool_tip="The number of iterations of the algorithm to preform. Increase if not detecting all rois but they are present in the eigen norm image",
                                        min=1, max=10000, step=1),

                               # IntInput(display_name="Number of time steps:",
                               #          program_name="total_num_time_steps",
                               #          on_change_function=lambda x,
                               #                                    y: data_handler.change_box_param(
                               #              x, y),
                               #          default_val=data_handler.box_params[
                               #              "total_num_time_steps"],
                               #          tool_tip=
                               #          "Number of time steps to break" +
                               #          "the processing into",
                               #          min=1, max=10000, step=1),
                               #
                               ]
                              )
# TODO add fill holes/number rois per spatial box to inputs



from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure
from qtpy.QtWidgets import *


class MplCanvas(FigureCanvasQTAgg):

    def __init__(self, parent=None, width=5, height=4, dpi=100):
        self.fig = Figure(figsize=(width, height), dpi=dpi, edgecolor=(1, 1, 1))
        self.fig.patch.set_facecolor((.098, .13725, .17647))
        self.axes = self.fig.add_subplot()
        self.fig.tight_layout(pad=0)
        self.axes.tick_params(colors='white')
        super(MplCanvas, self).__init__(self.fig)


class GraphItemPColor(QWidget):
    def __init__(self, *, data, x_label, y_label, x_ticks, y_ticks,
                 display_y_axis_ticks=True, display_axis_labels=True, aspect_ratio_x=9,
                 aspect_ratio_y=1):
        super(GraphItemPColor, self).__init__()
        self.setStyleSheet("""
                                QLabel {
                                background-color: rgba(0,0,0,0%)
                                }""")
        layout = QVBoxLayout()
        layout_h = QHBoxLayout()
        self.data = data
        self.y_label = y_label
        if len(data.shape) == 1:
            data = data.reshape((-1, 1))
            data = data.transpose((1, 0))
        self.graph = MplCanvas(width=aspect_ratio_x, height=aspect_ratio_y)
        self.graph.axes.pcolor(data, cmap='jet')

        # self.graph.axes.xticks(ticks=list(range(data.shape[0])), labels=x_ticks)
        if display_y_axis_ticks and y_ticks != None:
            self.graph.axes.yaxis.set_ticks(ticks=list(range(data.shape[0])))
            self.graph.axes.set_yticklabels(labels=y_ticks)
        if not display_y_axis_ticks:
            self.graph.axes.yaxis.set_ticks([])
        self.graph.axes.set_xlabel(x_label, color="white")
        # self.graph.axes.set_ylabel(y_label, color="white")
        layout_h.addWidget(QLabel(y_label))
        # test = QLabel("")
        # test.setMinimumHeight(200)
        # layout_h.addWidget(test)
        layout_h.addWidget(self.graph)

        layout.addLayout(layout_h, stretch=20)
        # layout.addWidget(QLabel(x_label), alignment=QtCore.Qt.AlignCenter,stretch=1)

        self.setLayout(layout)
        # self.setMinimumHeight(200)
        layout.setContentsMargins(0, 0, 10, 0)
        layout_h.setContentsMargins(0, 0, 00, 0)



import json
import json
import logging
from functools import reduce
from math import ceil

import dask
import numpy as np
import zarr
from PIL import Image
from dask import delayed
from skimage import feature
from tifffile import tifffile

from CIDAN.LSSC.SpatialBox import SpatialBox
from CIDAN.LSSC.functions.data_manipulation import load_filter_tif_stack, filter_stack, \
    reshape_to_2d_over_time, pixel_num_to_2d_cord, applyPCA
from CIDAN.LSSC.functions.eigen import loadEigenVectors
from CIDAN.LSSC.functions.pickle_funcs import *
from CIDAN.LSSC.functions.progress_bar import printProgressBarFilter, printProgressBar
from CIDAN.LSSC.functions.roi_extraction import roi_extract_image, combine_rois
from CIDAN.LSSC.functions.roi_filter import filterRoiList
from CIDAN.LSSC.functions.spatial_footprint import classify_components_ep
from CIDAN.LSSC.process_data import process_data
from CIDAN.TimeTrace.deltaFOverF import calculateDeltaFOverF
from CIDAN.TimeTrace.mean import calculateMeanTrace
from CIDAN.TimeTrace.waveletDenoise import waveletDenoise

logger1 = logging.getLogger("CIDAN.DataHandler")


class DataHandler:
    """
    Interacts with the algorithm and stores the current image data

    This class is initialized whenever data is loaded into the program, it handles
    saving the parameter file, running each of the algorithms in timetrace and LSSC.

    Attributes
    ----------
    global_params : Dict
        This saves the general params of for the dataset, basically whether parameters
        have changed so that the dataset, filters, boxes, eigen vectors, or
        roi extraction need to be recalculated


    """
    _global_params_default = {
        "save_intermediate_steps": True,
        "need_recalc_dataset_params": True,
        "need_recalc_filter_params": True,
        "need_recalc_box_params": True,
        "need_recalc_eigen_params": True,
        "need_recalc_roi_extraction_params": True,
        "load_into_mem": False,
        "num_threads": 8
    }

    _dataset_params_default = {
        "dataset_folder_path": "",
        "trials_loaded": [],
        "trials_all": [],
        "single_file_mode": False,
        "original_folder_trial_split": "",
        "slice_stack": False,
        "slice_every": 3,
        "slice_start": 0,
        "crop_stack": False,
        "crop_x": [0, 0],
        "crop_y": [0, 0],
        "trial_split": False,
        "trial_length": 400
    }

    _filter_params_default = {
        "median_filter": False,
        "median_filter_size": 3,
        "z_score": False,
        "hist_eq": False,
        "localSpatialDenoising": True,
        "pca": False,
        "pca_threshold": .97

    }
    _box_params_default = {
        "total_num_time_steps": 1,
        "total_num_spatial_boxes": 1,
        "spatial_overlap": 40
    }
    _eigen_params_default = {
        "eigen_vectors_already_generated": False,
        "num_eig": 50,
        "normalize_w_k": 25,
        "metric": "l2",
        "knn": 32,
        "accuracy": 75,
        "eigen_accuracy": 5,
        "connections": 40

    }
    _roi_extraction_params_default = {
        "elbow_threshold_method": True,
        "elbow_threshold_value": 1,
        "eigen_threshold_method": True,
        "eigen_threshold_value": .1,
        "num_eigen_vector_select": 5,
        "merge_temporal_coef": .90,
        "roi_size_min": 30,
        "roi_size_max": 600,
        "merge": True,
        "num_rois": 60,
        "fill_holes": True,
        "refinement": True,
        "max_iter": 100,
        "roi_circ_threshold": 20
    }
    _time_trace_params_default = {
        "time_trace_type": "Mean",
        "denoise": True
    }

    def __init__(self, data_path, save_dir_path, save_dir_already_created, trials=[],
                 parameter_file=False, load_into_mem=True):
        """
        Initializes the object
        Parameters
        ----------
        data_path : str
            This is the path towards the folder containing the data files/folder null
            if save dir is already created
        save_dir_path : str
            save folder location
        save_dir_already_created : bool
            whether to load all the info from the save dir, if this is true data_path
            and trials don't matter
        trials : List[str]
            The list of either names of folders that contain many tiff files or a
            list of tiff files
        """
        self.parameter_file_name = "parameters.json"
        # TODO add loaded trials and all trials parameter here
        # TODO make sure if trial list includes files that aren't valid it works
        self.color_list = [(218, 67, 34),
                           (132, 249, 22), (22, 249, 140), (22, 245, 249),
                           (22, 132, 249), (224, 22, 249), (249, 22, 160)]

        self.save_dir_path = save_dir_path

        self.rois_loaded = False  # whether roi variables have been created
        if parameter_file:
            self.parameter_file_name = os.path.basename(parameter_file)
            self.save_dir_path = os.path.dirname(parameter_file)
            self.create_new_save_dir()
            valid = self.load_param_json()
            self.global_params = DataHandler._global_params_default.copy()
            # these indices are used for which trials to use for roi extract
            self._trials_loaded_indices = [num for num, x in enumerate(self.trials_all)
                                           if x in self.trials_loaded]
            # these are used to determine which trials to calculate time_traces for
            self.trials_loaded_time_trace_indices = [num for num, x in
                                                     enumerate(self.trials_all)
                                                     if x in self.trials_loaded]


        elif save_dir_already_created:  # this loads everything from the save dir
            valid = self.load_param_json()
            self.time_trace_params = DataHandler._time_trace_params_default.copy()
            if self.dataset_params["single_file_mode"]:
                if not os.path.isdir(
                        os.path.join(self.save_dir_path, "temp_files/dataset.zarr")):
                    self.transform_data_to_zarr()
            # time trace params are not currently saved

            # these indices are used for which trials to use for roi extract
            self._trials_loaded_indices = [num for num, x in enumerate(self.trials_all)
                                           if x in self.trials_loaded]
            # these are used to determine which trials to calculate time_traces for
            self.trials_loaded_time_trace_indices = [num for num, x in
                                                     enumerate(self.trials_all)
                                                     if x in self.trials_loaded]
            # this loads the dataset and calculate the specified filters
            if not self.load_into_mem:
                try:
                    self.load_data()
                    print("Loaded Previous Session")
                except:
                    self.__delattr__("dataset_trials_filtered")
                    temp = self.global_params["need_recalc_box_params"]
                    self.calculate_filters()
                    self.global_params["need_recalc_box_params"] = temp
            else:
                temp = self.global_params["need_recalc_box_params"]
                self.calculate_filters()
                self.global_params["need_recalc_box_params"] = temp
            # if there are ROIs saved in the save dir load them and calculate time
            # traces

            if self.rois_exist:
                try:
                    self.load_rois()
                    self.calculate_time_traces()
                    self.rois_loaded = True
                except:
                    print("ROI loading Failed")
            if not valid:
                raise FileNotFoundError("Save directory not valid")
        else:
            # start with the default values specified in this class

            self.global_params = DataHandler._global_params_default.copy()
            self.global_params["load_into_mem"] = load_into_mem
            self.dataset_params = DataHandler._dataset_params_default.copy()
            self.filter_params = DataHandler._filter_params_default.copy()
            self.box_params = DataHandler._box_params_default.copy()
            self.box_params_processed = DataHandler._box_params_default.copy()
            self.roi_extraction_params = DataHandler._roi_extraction_params_default.copy()
            self.time_trace_params = DataHandler._time_trace_params_default.copy()

            self.dataset_params["dataset_folder_path"] = data_path
            self.dataset_params["trials_loaded"] = trials
            self.trials_loaded = trials
            if not os.path.isdir(
                    os.path.join(data_path, trials[0])) and len(
                self.trials_loaded) == 1:
                self.dataset_params["single_file_mode"] = True
            if len(self.trials_loaded) == 1:
                self.dataset_params["trial_split"] = True
                self.dataset_params["original_folder_trial_split"] = trials[0]
            # these are all files in the data_path directory if the user wants to
            # calculate time traces than for more trials than they originally loaded
            self.trials_all = sorted(os.listdir(data_path))
            if not os.path.isdir(os.path.join(data_path, trials[0])):
                self.trials_all = [x for x in self.trials_all if ".tif" in x]
            self.dataset_params["trials_all"] = self.trials_all

            self.box_params["total_num_time_steps"] = len(trials)
            self.eigen_params = DataHandler._eigen_params_default.copy()

            valid = self.create_new_save_dir()
            if not valid:
                raise FileNotFoundError("Please chose an empty directory for your " +
                                        "save directory")
            if self.dataset_params["single_file_mode"]:
                self.transform_data_to_zarr()
            self.time_traces = []
            self._trials_loaded_indices = [num for num, x in enumerate(self.trials_all)
                                           if x in self.trials_loaded]
            self.trials_loaded_time_trace_indices = [num for num, x in
                                                     enumerate(self.trials_all)
                                                     if x in self.trials_loaded]
            self.save_new_param_json()

    def __del__(self):
        try:
            for x in list(self.__dict__.items()):
                self.__dict__[x[0]] = None
        except TypeError:
            pass

    @property
    def dataset_trials_filtered_loaded(self):
        """
        Returns the filtered data that is currently loaded
        """
        return [self.dataset_trials_filtered[x] for x in self._trials_loaded_indices]

    @property
    def dataset_trials_loaded(self):
        """
        Returns the unfiltered data that is currently loaded
        """
        return [self.dataset_trials[x] for x in self._trials_loaded_indices]

    @property
    def param_path(self):
        """
        Returns the path to the parameter file
        """
        return os.path.join(self.save_dir_path, self.parameter_file_name)

    @property
    def eigen_vectors_exist(self):
        """
        Returns if the eigen vectors are saved for the current settings
        """

        eigen_dir = os.path.join(self.save_dir_path, "eigen_vectors")

        file_names = [
            "eigen_vectors_box_{}_{}.pickle".format(spatial_box_num, time_box_num)
            for spatial_box_num in range(self.box_params["total_num_spatial_boxes"])
            for time_box_num in range(self.box_params["total_num_time_steps"])]
        return all(pickle_exist(x, output_directory=eigen_dir) for x in file_names)

    @property
    def rois_exist(self):
        """
        Return if the roi save file exists
        """
        return not (self.global_params["need_recalc_eigen_params"] or
                    self.global_params[
                        "need_recalc_roi_extraction_params"] or self.global_params[
                        "need_recalc_box_params"]) and pickle_exist("rois",
                                                                    output_directory=self.save_dir_path)

    @property
    def load_into_mem(self):
        return self.global_params["load_into_mem"]
    def load_rois(self):
        """
        Loads the ROIs if they exist and then generates the other variables associated
        with them
        """
        if pickle_exist("rois", output_directory=self.save_dir_path):
            self.rois = pickle_load("rois", output_directory=self.save_dir_path)
            self.gen_roi_display_variables()

    def save_rois(self, rois):
        """
        Saves the rois to files
        Parameters
        ----------
        rois : List[np.ndarray]
            A list pixels of each roi, each pixel is its 1d cordinate

        """
        if os.path.isdir(self.save_dir_path):
            pickle_save(rois, "rois", output_directory=self.save_dir_path)

    def load_param_json(self):
        """
        Loads the parameter json file and saves it to all the parameter values
        """

        try:
            with open(self.param_path, "r") as f:
                all_params = json.loads(f.read())
            self.global_params = all_params["global_params"]
            self.dataset_params = all_params["dataset_params"]
            self.filter_params = all_params["filter_params"]
            self.box_params = all_params["box_params"]
            self.box_params_processed = all_params["box_params"].copy()
            self.eigen_params = all_params["eigen_params"]
            self.time_trace_params = all_params["time_trace_params"]
            self.roi_extraction_params = all_params["roi_extraction_params"]
            self.trials_loaded = self.dataset_params["trials_loaded"]
            self.trials_all = self.dataset_params["trials_all"]
            return True
        except KeyError:
            raise KeyError("Please Choose a valid parameter file")
        except FileNotFoundError:
            raise FileNotFoundError("Can't find parameter file")
        except NameError:
            raise FileNotFoundError("Can't find parameter file")

    def save_new_param_json(self):
        """
        Saves the parameters to the parameter file
        """
        try:
            with open(self.param_path, "w") as f:
                all_params = {
                    "global_params": self.global_params,
                    "dataset_params": self.dataset_params,
                    "filter_params": self.filter_params,
                    "box_params": self.box_params_processed,
                    "eigen_params": self.eigen_params,
                    "roi_extraction_params": self.roi_extraction_params,
                    "time_trace_params": self.time_trace_params
                }
                f.truncate(0)
                f.write(json.dumps(all_params))
        except:
            raise FileNotFoundError("Error saving parameters, please restart software")
        pass

    def create_new_save_dir(self):
        """
        Creates a new save directory
        """
        try:
            if not os.path.isdir(self.save_dir_path):
                os.mkdir(self.save_dir_path)
            eigen_vectors_folder_path = os.path.join(self.save_dir_path,
                                                     "eigen_vectors/")

            if not os.path.isdir(eigen_vectors_folder_path):
                os.mkdir(eigen_vectors_folder_path)
            embedding_images_path = os.path.join(self.save_dir_path,
                                                 "embedding_norm_images/")
            if not os.path.isdir(embedding_images_path):
                os.mkdir(embedding_images_path)
            eigen_vectors_folder_path = os.path.join(self.save_dir_path,
                                                     "temp_files/")

            if not os.path.isdir(eigen_vectors_folder_path):
                os.mkdir(eigen_vectors_folder_path)
            return True
        except:
            raise FileNotFoundError("Couldn't create folder please try again")

    def change_global_param(self, param_name, new_value):
        """
        Used to change a param in global params area.
        Parameters
        ----------
        param_name : str
            name of parameter to change
        new_value : ?
            the new value to the parameter

        Returns
        -------
        True if successful
        """
        if param_name in self.global_params:
            self.global_params[param_name] = new_value
            self.save_new_param_json()
            return True
        else:
            return False

    def change_dataset_param(self, param_name, new_value):
        """
        Used to change a param in dataset params area.
        Parameters
        ----------
        param_name : str
            name of parameter to change
        new_value : ?
            the new value to the parameter

        Returns
        -------
        True if successful
        """
        if param_name in self.dataset_params:
            self.dataset_params[param_name] = new_value
            self.global_params["need_recalc_dataset_params"] = True
            self.global_params["need_recalc_box_params"] = True
            self.global_params["need_recalc_eigen_params"] = True
            # self.save_new_param_json()
            return True
        else:
            return False

    def change_filter_param(self, param_name, new_value):
        """
        Used to change a param in filter params area.
        Parameters
        ----------
        param_name : str
            name of parameter to change
        new_value : ?
            the new value to the parameter

        Returns
        -------
        True if successful
        """
        if param_name in self.filter_params:
            self.filter_params[param_name] = new_value
            self.global_params["need_recalc_filter_params"] = True
            self.global_params["need_recalc_box_params"] = True

            self.global_params["need_recalc_eigen_params"] = True
            # self.save_new_param_json()
            return True
        else:
            return False

    def change_box_param(self, param_name, new_value):
        """
        Used to change a param in box params area.
        Parameters
        ----------
        param_name : str
            name of parameter to change
        new_value : ?
            the new value to the parameter

        Returns
        -------
        True if successful
        """
        if param_name in self.box_params:
            # if param_name == "total_num_spatial_boxes":
            #     assert (int(new_value**.5))**2 == new_value, "Please make sure Number of Spatial Boxes is a square number"
            self.box_params[param_name] = new_value
            self.global_params["need_recalc_box_params"] = True
            self.global_params["need_recalc_eigen_params"] = True
            self.global_params["need_recalc_roi_extraction_params"] = True
            # self.save_new_param_json()
            return True
        else:
            return False

    def change_eigen_param(self, param_name, new_value):
        """
        Used to change a param in eigen params area.
        Parameters
        ----------
        param_name : str
            name of parameter to change
        new_value : ?
            the new value to the parameter

        Returns
        -------
        True if successful
        """
        if param_name in self.eigen_params:
            self.eigen_params[param_name] = new_value
            self.global_params["need_recalc_eigen_params"] = True
            # self.save_new_param_json()
            return True
        else:
            return False

    def change_roi_extraction_param(self, param_name, new_value):
        """
        Used to change a param in roi_extraction params area.
        Parameters
        ----------
        param_name : str
            name of parameter to change
        new_value : ?
            the new value to the parameter

        Returns
        -------
        True if successful
        """
        if param_name in self.roi_extraction_params:
            self.roi_extraction_params[param_name] = new_value
            self.global_params["need_recalc_roi_extraction_params"] = True
            # self.save_new_param_json()
            return True
        else:
            return False

    def transform_data_to_zarr(self):
        image = tifffile.imread(os.path.join(self.dataset_params["dataset_folder_path"],
                                             self.dataset_params[
                                                 "original_folder_trial_split"]))
        z1 = zarr.open(os.path.join(self.save_dir_path,
                                    'temp_files/dataset.zarr'), mode='w',
                       shape=image.shape,
                       chunks=(32, 128, 128), dtype=np.float32)
        z1[:] = image

    def calculate_dataset(self, ) -> np.ndarray:
        """
        Loads each trial, applying crop and slicing, sets them to self.dataset_trials

        """
        # TODO make it so it does't load the dataset every time

        dataset_trials = [False] * len(self.trials_all)
        for trial_num in self._trials_loaded_indices:
            dataset_trials[trial_num] = self.load_trial_dataset_step(trial_num)
        dataset_trials = dask.compute(*dataset_trials)
        self.global_params["need_recalc_dataset_params"] = False
        self.shape = [dataset_trials[self._trials_loaded_indices[0]].shape[1],
                      dataset_trials[self._trials_loaded_indices[0]].shape[2]]
        print(dataset_trials[0])
        if self.dataset_params["crop_x"][1] == 0:
            self.dataset_params["crop_x"][1] = self.shape[0]
            self.dataset_params["crop_y"][1] = self.shape[1]
        print("Finished Calculating Dataset")
        return dataset_trials

    @delayed
    def load_trial_dataset_step(self, trial_num):
        """
        Delayed function step that loads a single trial
        Parameters
        ----------
        trial_num : int
            trial num to load in trials_all

        Returns
        -------
        Trial as a np.ndarray
        """
        total_size, stack = load_filter_tif_stack(
            path=os.path.join(self.dataset_params["dataset_folder_path"],
                              self.trials_all[trial_num] if not self.dataset_params[
                                  "trial_split"] else self.dataset_params[
                                  "original_folder_trial_split"]),
            filter=False,
            median_filter=False,
            median_filter_size=(1, 3, 3),
            z_score=False,
            slice_stack=self.dataset_params[
                "slice_stack"],
            slice_start=self.dataset_params[
                "slice_start"],

            slice_every=self.dataset_params[
                "slice_every"],
            crop_stack=self.dataset_params[
                "crop_stack"],
            crop_x=self.dataset_params[
                "crop_x"],
            crop_y=self.dataset_params[
                "crop_y"], load_into_mem=self.load_into_mem,
            trial_split=self.dataset_params["trial_split"],
            trial_split_length=self.dataset_params["trial_length"], trial_num=trial_num,
            zarr_path=False if not self.dataset_params["single_file_mode"] else
            os.path.join(self.save_dir_path, "temp_files/dataset.zarr")
        )
        self.total_size = total_size
        return stack

    @delayed
    def load_trial_filter_step(self, trial_num, dataset=False, loaded_num=False):
        """
        Delayed function step that applies filter to a single trial
        Parameters
        ----------
        trial_num : int
            trial num to load in trials_all
        dataset : ndarray
            Dataset to process if false load dataset

        Returns
        -------
        Filtered trial as a np.ndarray
        """
        if not self.load_into_mem:
            cur_stack = filter_stack(

                stack=dataset if type(
                    dataset) != bool else self.load_trial_dataset_step(
                    trial_num).compute(),
                median_filter_size=(self.filter_params[
                                        "median_filter_size"],
                                    self.filter_params[
                                        "median_filter_size"],
                                    self.filter_params[
                                        "median_filter_size"]),
                median_filter=self.filter_params[
                    "median_filter"],
                z_score=self.filter_params["z_score"],
                hist_eq=self.filter_params["hist_eq"],
                localSpatialDenoising=self.filter_params["localSpatialDenoising"])
            z1 = zarr.open(os.path.join(self.save_dir_path,
                                        'temp_files/%s.zarr' % self.trials_all[
                                            trial_num]), mode='w',
                           shape=cur_stack.shape,
                           chunks=(32,128,128), dtype=np.float32)
            z1[:] = cur_stack
            if type(loaded_num) != bool:
                self.mean_images[loaded_num] = np.mean(cur_stack, axis=0)
                self.max_images[loaded_num] = np.max(cur_stack, axis=0)


        else:
            cur_stack = filter_stack(

                stack=dataset if type(
                    dataset) != bool else self.load_trial_dataset_step(
                    trial_num).compute(),
                median_filter_size=(self.filter_params[
                                        "median_filter_size"],
                                    self.filter_params[
                                        "median_filter_size"],
                                    self.filter_params[
                                        "median_filter_size"]),
                median_filter=self.filter_params[
                    "median_filter"],
                z_score=self.filter_params["z_score"],
                hist_eq=self.filter_params["hist_eq"])

            if type(loaded_num) != bool:
                self.mean_images[loaded_num] = np.mean(cur_stack, axis=0)
                self.max_images[loaded_num] = np.max(cur_stack, axis=0)
        if self.filter_params["pca"] and type(loaded_num) != bool:
            pca = applyPCA(cur_stack, self.filter_params["pca_threshold"])
            if self.load_into_mem:
                self.pca_decomp[loaded_num] = pca
            else:
                z2 = zarr.open(os.path.join(self.save_dir_path,
                                            'temp_files/%s_pca.zarr' % self.trials_all[
                                                trial_num]), mode='w',
                               shape=pca.shape,
                               chunks=(3, 64, 64))
                z2[:] = pca
                self.pca_decomp[loaded_num] = z2
        with open(os.path.join(self.save_dir_path,
                               'temp_files/filter/%s' % self.trials_all[
                                   trial_num]), "w") as f:
            f.write("done")
        printProgressBarFilter(
            total_num_spatial_boxes=self.box_params["total_num_spatial_boxes"],
            total_num_time_steps=self.box_params["total_num_time_steps"],
            save_dir=self.save_dir_path, progress_signal=self.progress_signal)

        if self.load_into_mem:
            return cur_stack
        else:
            return z1

    def calculate_filters(self, progress_signal=None):
        """
        Applies filter to each trial, sets them to self.dataset_trials_filtered

        Returns
        -------
        A list of filtered trials
        """
        if self.global_params["need_recalc_filter_params"] or self.global_params[
            "need_recalc_dataset_params"] or \
                not hasattr(self, "dataset_trials_filtered"):
            self.progress_signal = progress_signal
            print("Started Calculating Filters")

            self.update_trial_list()
            save_dir = self.save_dir_path
            if not os.path.isdir(os.path.join(save_dir, "temp_files/filter")):
                os.mkdir(os.path.join(save_dir, "temp_files/filter"))
            filelist = [f for f in
                        os.listdir(os.path.join(save_dir, "temp_files/filter"))]
            for f in filelist:
                os.remove(
                    os.path.join(os.path.join(save_dir, "temp_files/filter"), f))

            printProgressBarFilter(
                total_num_spatial_boxes=self.box_params["total_num_spatial_boxes"],
                total_num_time_steps=self.box_params["total_num_time_steps"],
                save_dir=self.save_dir_path, progress_signal=self.progress_signal)

            self.dataset_trials_filtered = [False] * len(self.trials_all)
            if self.filter_params["pca"]:
                self.pca_decomp = [False] * len(self._trials_loaded_indices)
            self.max_images = [False] * len(self._trials_loaded_indices)
            self.mean_images = [False] * len(self._trials_loaded_indices)
            if not self.load_into_mem:
                for num, trial_num in enumerate(self._trials_loaded_indices):
                    self.dataset_trials_filtered[trial_num] = self.load_trial_filter_step(
                        trial_num, self.load_trial_dataset_step(trial_num), loaded_num = num)
                    if num % 3 == 0:
                        self.dataset_trials_filtered = list(
                            dask.compute(*self.dataset_trials_filtered))

                self.dataset_trials_filtered = list(
                    dask.compute(*self.dataset_trials_filtered))
            else:
                for num, trial_num in enumerate(self._trials_loaded_indices):
                    self.dataset_trials_filtered[trial_num] = self.load_trial_filter_step(
                        trial_num, self.load_trial_dataset_step(trial_num), loaded_num=num)
                self.dataset_trials_filtered = list(
                    dask.compute(*self.dataset_trials_filtered))
            self.shape = [
                self.dataset_trials_filtered[self._trials_loaded_indices[0]].shape[1],
                self.dataset_trials_filtered[self._trials_loaded_indices[0]].shape[2]]
            if self.dataset_params["crop_x"][1] == 0:
                self.dataset_params["crop_x"][1] = self.shape[0]
                self.dataset_params["crop_y"][1] = self.shape[1]
            mean_image_stack = np.dstack(self.mean_images)
            with open(os.path.join(self.save_dir_path,
                                   'temp_files/filter/mean'), "w") as f:
                f.write("done")
            printProgressBarFilter(
                total_num_spatial_boxes=self.box_params["total_num_spatial_boxes"],
                total_num_time_steps=self.box_params["total_num_time_steps"],
                save_dir=self.save_dir_path, progress_signal=self.progress_signal)

            max_image_stack = np.dstack(self.max_images)
            with open(os.path.join(self.save_dir_path,
                                   'temp_files/filter/max'), "w") as f:
                f.write("done")
            mean_image_zarr = zarr.open(os.path.join(self.save_dir_path,
                                                     'temp_files/mean.zarr'), mode='w',
                                        shape=mean_image_stack.shape,
                                        chunks=(40, 40, 1))
            mean_image_zarr[:] = mean_image_stack

            max_image_zarr = zarr.open(os.path.join(self.save_dir_path,
                                                    'temp_files/max.zarr'), mode='w',
                                       shape=mean_image_stack.shape,
                                       chunks=(40, 40, 1))
            max_image_zarr[:] = max_image_stack
            printProgressBarFilter(
                total_num_spatial_boxes=self.box_params["total_num_spatial_boxes"],
                total_num_time_steps=self.box_params["total_num_time_steps"],
                save_dir=self.save_dir_path, progress_signal=self.progress_signal)

            # self.mean_images = [np.mean(x[:], axis=0) for x in
            #                     self.dataset_trials_filtered_loaded]
            #
            # self.max_images = [np.max(x[:], axis=0) for x in
            #                    self.dataset_trials_filtered_loaded]

            # self.temporal_correlation_image = calculate_temporal_correlation(self.dataset_filtered)
            self.global_params["need_recalc_filter_params"] = False
            self.global_params["need_recalc_dataset_params"] = False

            # self.global_params["need_recalc_box_params"] = True
            self.save_new_param_json()
            self.delete_roi_vars()
            # self.global_params["need_recalc_eigen_params"] = True
        return self.dataset_trials_filtered

    @property
    def real_trials(self):
        return not (self.dataset_params["single_file_mode"] or self.dataset_params[
            "trial_split"] or len(self.trials_loaded) == 1)
    def load_data(self):
        self.update_trial_list()
        self.dataset_trials_filtered = [False] * len(self.trials_all)
        if self.filter_params["pca"]:
            self.pca_decomp = [False] * len(self.trials_loaded_time_trace_indices)
        for trial_num in self._trials_loaded_indices:
            self.dataset_trials_filtered[trial_num] = zarr.open(
                os.path.join(self.save_dir_path,
                             'temp_files/%s.zarr' % self.trials_all[trial_num]),
                mode="r")
            if self.filter_params["pca"]:
                self.pca_decomp[trial_num] = zarr.open(
                    os.path.join(self.save_dir_path,
                                 'temp_files/%s_pca.zarr' % self.trials_all[trial_num]),
                    mode="r")
        max_image_zarr = zarr.open(
            os.path.join(self.save_dir_path,
                         'temp_files/max.zarr'),
            mode="r")
        self.max_images = [False] * len(self.trials_loaded_time_trace_indices)
        mean_image_zarr = zarr.open(
            os.path.join(self.save_dir_path,
                         'temp_files/mean.zarr'),
            mode="r")
        self.mean_images = [False] * len(self.trials_loaded_time_trace_indices)
        for x in range(len(self._trials_loaded_indices)):
            self.max_images[x] = max_image_zarr[:, :, x]
            self.mean_images[x] = mean_image_zarr[:, :, x]
        self.shape = [
            self.dataset_trials_filtered[self._trials_loaded_indices[0]].shape[1],
            self.dataset_trials_filtered[self._trials_loaded_indices[0]].shape[2]]
        if self.dataset_params["crop_x"][1] == 0:
            self.dataset_params["crop_x"][1] = self.shape[0]
            self.dataset_params["crop_y"][1] = self.shape[1]
        # self.mean_images = [np.mean(x[:], axis=0) for x in
        #                     self.dataset_trials_filtered_loaded]
        #
        # self.max_images = [np.max(x[:], axis=0) for x in
        #                    self.dataset_trials_filtered_loaded]
        self.global_params["need_recalc_filter_params"] = False
        self.global_params["need_recalc_dataset_params"] = False

    def update_trial_list(self):
        if self.dataset_params["trial_split"] and self.dataset_params[
            "original_folder_trial_split"] != "":
            if self.dataset_params["single_file_mode"]:
                z1 = zarr.open(
                    os.path.join(self.save_dir_path, "temp_files/dataset.zarr"),
                    mode="r")
                self.trials_loaded = [str(x) for x in range(ceil(z1.shape[0] /
                                                                 self.dataset_params[
                                                                     "trial_length"]))]
            else:
                self.trials_loaded = [str(x) for x in range(ceil(len(os.listdir(
                os.path.join(self.dataset_params["dataset_folder_path"],
                             self.dataset_params["original_folder_trial_split"]))) /
                                                             self.dataset_params[
                                                                 "trial_length"]))]
            self.trials_all = self.trials_loaded.copy()
            self.box_params["total_num_time_steps"] = len(self.trials_loaded)
        elif self.dataset_params["original_folder_trial_split"] != "":
            self.trials_loaded = [
                self.dataset_params["original_folder_trial_split"]]
            self.trials_all = [self.dataset_params["original_folder_trial_split"]]
            self.box_params["total_num_time_steps"] = len(self.trials_loaded)
        if self.dataset_params["original_folder_trial_split"] != "":
            self._trials_loaded_indices = [num for num, x in
                                           enumerate(self.trials_all)
                                           if x in self.trials_loaded]
            self.trials_loaded_time_trace_indices = [num for num, x in
                                                     enumerate(self.trials_all)
                                                     if x in self.trials_loaded]

    def calculate_roi_extraction(self, progress_signal=None):
        """
        Extracts Rois and sets them to self.rois
        """
        if self.global_params["need_recalc_eigen_params"] or self.global_params[
            "need_recalc_roi_extraction_params"] or self.global_params[
            "need_recalc_box_params"] or self.global_params[
            "need_recalc_dataset_params"] or \
                self.global_params["need_recalc_filter_params"]:
            assert (int(
                self.box_params[
                    "total_num_spatial_boxes"] ** .5)) ** 2 == self.box_params[
                       "total_num_spatial_boxes"], "Please make sure Number of Spatial Boxes is a square number"
            try:
                self.calculate_filters()
                eigen_need_recalc = self.global_params["need_recalc_eigen_params"] or self.global_params[
                    "need_recalc_box_params"]
                self.global_params["need_recalc_eigen_params"] = False
                self.global_params[
                    "need_recalc_roi_extraction_params"] = False
                temp_params = self.box_params.copy()
                self.global_params[
                    "need_recalc_box_params"] = False
                self.rois = process_data(num_threads=self.global_params[
                    "num_threads"], test_images=False, test_output_dir="",
                                         save_dir=self.save_dir_path,
                                         save_intermediate_steps=self.global_params[
                                             "save_intermediate_steps"],
                                         image_data=self.dataset_trials_filtered_loaded,
                                         eigen_vectors_already_generated=(not
                                                                          eigen_need_recalc) and
                                                                         self.global_params[
                                                                             "save_intermediate_steps"] and self.eigen_vectors_exist,
                                         save_embedding_images=True,
                                         total_num_time_steps=self.box_params[
                                             "total_num_time_steps"],
                                         total_num_spatial_boxes=self.box_params[
                                             "total_num_spatial_boxes"],
                                         spatial_overlap=self.box_params[
                                                             "spatial_overlap"] // 2,

                                         metric=self.eigen_params["metric"],
                                         knn=self.eigen_params["knn"],
                                         accuracy=self.eigen_params["accuracy"],
                                         eigen_accuracy=self.eigen_params[
                                             "eigen_accuracy"],
                                         connections=self.eigen_params[
                                             "connections"],
                                         normalize_w_k=self.eigen_params[
                                             "normalize_w_k"],
                                         num_eig=self.eigen_params["num_eig"],
                                         merge=self.roi_extraction_params["merge"],
                                         num_rois=self.roi_extraction_params[
                                             "num_rois"],
                                         refinement=self.roi_extraction_params[
                                             "refinement"],
                                         num_eigen_vector_select=
                                         self.roi_extraction_params[
                                             "num_eigen_vector_select"],
                                         max_iter=self.roi_extraction_params[
                                             "max_iter"],
                                         roi_size_min=self.roi_extraction_params[
                                             "roi_size_min"],
                                         fill_holes=self.roi_extraction_params[
                                             "fill_holes"],

                                         elbow_threshold_method=
                                         self.roi_extraction_params[
                                             "elbow_threshold_method"],
                                         elbow_threshold_value=
                                         self.roi_extraction_params[
                                             "elbow_threshold_value"],
                                         eigen_threshold_method=
                                         self.roi_extraction_params[
                                             "eigen_threshold_method"],
                                         eigen_threshold_value=
                                         self.roi_extraction_params[
                                             "eigen_threshold_value"],
                                         merge_temporal_coef=
                                         self.roi_extraction_params[
                                             "merge_temporal_coef"],
                                         roi_size_max=self.roi_extraction_params[
                                             "roi_size_max"],
                                         pca=self.filter_params["pca"],
                                         pca_data= self.pca_decomp if self.filter_params[
                                             "pca"] else False,
                                         progress_signal=progress_signal)
                self.box_params_processed = temp_params
                self.save_new_param_json()
                roi_valid_list = filterRoiList(self.rois, self.shape)
                self.rois = [x for x, y in zip(self.rois, roi_valid_list) if
                             y >= self.roi_extraction_params[
                                 "roi_circ_threshold"]]
                self.save_rois(self.rois)
                print("Calculating Time Traces")
                self.gen_roi_display_variables()
                self.calculate_time_traces()

                self.rois_loaded = True
            except FileNotFoundError as e:
                self.global_params["need_recalc_eigen_params"] = False
                logger1.error(e)
                print(
                    "Please try again there was an internal error in the roi extraction process")
                raise AssertionError()

    def gen_roi_display_variables(self):
        """
        Generates the other variables associated with displaying ROIs
        """
        self.roi_circ_list = filterRoiList(self.rois, self.shape,
                                           self.roi_extraction_params[
                                               "roi_circ_threshold"])
        roi_list_2d_cord = [
            pixel_num_to_2d_cord(x, volume_shape=self.shape) for x in
            self.rois]
        self.roi_max_cord_list = [np.max(x, axis=1) for x in roi_list_2d_cord]
        self.roi_min_cord_list = [np.min(x, axis=1) for x in roi_list_2d_cord]
        self.pixel_with_rois_flat = np.zeros(
            [self.shape[0] * self.shape[1]])
        self.pixel_with_rois_color_flat = np.zeros(
            [self.shape[0] * self.shape[1], 3])
        edge_roi_image = np.zeros([self.shape[0], self.shape[1]])
        for num, roi in enumerate(self.rois):
            cur_color = self.color_list[num % len(self.color_list)]
            self.pixel_with_rois_flat[roi] = num + 1
            self.pixel_with_rois_color_flat[roi] = cur_color
            roi_edge = np.zeros(
                [self.shape[0] * self.shape[1]])
            roi_edge[roi] = 255
            edge_roi_image += feature.canny(np.reshape(roi_edge,
                                                  [self.shape[0],
                                                   self.shape[1]]))

        edge_roi_image[edge_roi_image > 255] = 255
        self.edge_roi_image_flat = np.reshape(edge_roi_image, [-1, 1]) * 255
        self.pixel_with_rois_color = np.reshape(self.pixel_with_rois_color_flat,
                                                [self.shape[0],
                                                 self.shape[1], 3])
        try:
            self.eigen_norm_image = np.asarray(Image.open(
                os.path.join(self.save_dir_path,
                             "embedding_norm_images/embedding_norm_image.png")))
        except:
            print("Can't generate eigen Norm image please try again")
        self.roi_time_trace_need_update = []
        for _ in range(len(self.rois)):
            self.roi_time_trace_need_update.append(False)

    def calculate_time_traces(self, report_progress=None):
        """
        Calculates the time traces for every roi in self.rois
        """

        self.time_traces = []
        for _ in range(len(self.rois)):
            self.time_traces.append([])
            for _ in range(len(self.trials_all)):
                self.time_traces[-1].append(False)
        calc_list = []
        for trial_num in self.trials_loaded_time_trace_indices:
            data = self.load_trial_dataset_step(trial_num).compute()
            data_2d = reshape_to_2d_over_time(data[:])
            del data
            # if type(self.dataset_trials_filtered[trial_num]) == bool:
            #     data = self.load_trial_filter_step(
            #         trial_num).compute()
            #     self.dataset_trials_filtered[trial_num] = data
            #
            #     del data
            # else:
            #     data_2d = reshape_to_2d_over_time(
            #         self.dataset_trials_filtered[trial_num])[:]
            calc_list = []
            for roi in range(len(self.rois)):
                calc_list.append(
                    (delayed)(
                        self.calculate_time_trace(roi + 1, trial_num, data_2d=data_2d)))
            dask.compute(*calc_list)
            if report_progress is not None:
                printProgressBar(self.trials_loaded_time_trace_indices.index(trial_num),
                                 total=len(self.trials_loaded_time_trace_indices) + 2,
                                 prefix="Time Trace Calculation Progress:",
                                 suffix="Complete", progress_signal=report_progress)
        if not self.real_trials:
            for roi_num in range(len(self.time_traces)):
                self.time_traces[roi_num] = [np.hstack(
                    [self.time_traces[roi_num][x] for x in
                     self.trials_loaded_time_trace_indices])]
                if self.time_trace_params["time_trace_type"] == "DeltaF Over F":
                    self.time_traces[roi_num] = [
                        calculateDeltaFOverF(0, self.time_traces[roi_num][0],
                                             denoise=self.time_trace_params[
                                                 "denoise"], given_mean=True)]
                elif self.time_trace_params["denoise"]:
                    self.time_traces[roi_num] = [waveletDenoise(
                        self.time_traces[roi_num][0].reshape((1, -1))).reshape((-1))]

        self.roi_time_trace_need_update = [False for _ in
                                           self.roi_time_trace_need_update]

        if os.path.isdir(self.save_dir_path):
            pickle_save(self.time_traces,
                        "time_traces_{0}{1}".format(
                            self.time_trace_params["time_trace_type"],
                            "_denoised" if self.time_trace_params["denoise"] else ""),
                        output_directory=self.save_dir_path)
        self.rois_loaded = True

    def calculate_time_trace(self, roi_num, trial_num=None, data_2d=None):
        """
        Calculates a time trace for a certain ROI and save to time trace list
        Parameters
        ----------
        roi_num
            roi to calculate for this starts at [1..number of rois]
        trial_num
            indice of trial in trials_all starts at [0, number of trials-1] if
            none then calculates for all trials
        """
        trial_nums = [trial_num]
        if trial_num == None:
            trial_nums = self.trials_loaded_time_trace_indices
        for trial_num in trial_nums:
            roi = self.rois[roi_num - 1]

            if type(self.dataset_trials_filtered[
                        trial_num]) == bool and data_2d == None:
                self.dataset_trials_filtered[trial_num] = self.load_trial_filter_step(
                    trial_num).compute()
            if data_2d is None:
                data_2d = reshape_to_2d_over_time(
                    self.dataset_trials_filtered[trial_num])
            if self.time_trace_params[
                "time_trace_type"] == "DeltaF Over F" and self.real_trials:
                time_trace = calculateDeltaFOverF(roi, data_2d,
                                                  denoise=self.time_trace_params[
                                                      "denoise"] if self.real_trials else False)
            else:
                time_trace = calculateMeanTrace(roi, data_2d,
                                                denoise=self.time_trace_params[
                                                    "denoise"] if self.real_trials else False)
            self.time_traces[roi_num - 1][trial_num] = time_trace

    def get_time_trace(self, num, trial=None):
        """
        Returns the time trace for a certain roi over all currently selected trials
        Parameters
        ----------
        num : int
            ROI num starts at 1 to num rois

        Returns
        -------
        np.ndarray of the time trace
        """
        if not self.real_trials:
            if len(self.time_traces) > num - 1 and type(
                    self.time_traces[num - 1][0]) != bool:
                return self.time_traces[num - 1][0]
            else:
                return False
        if (trial == None):
            num = num - 1
            if self.real_trials:
                output = np.ndarray(shape=(0))
                for trial_num in self.trials_loaded_time_trace_indices:
                    if type(self.time_traces[num][trial_num]) == bool:
                        self.calculate_time_trace(num + 1, trial_num)
                    output = np.hstack([output, self.time_traces[num][trial_num]])
            else:
                return self.time_traces[num]
        else:
            num = num - 1
            if type(self.time_traces[num][trial]) == bool:
                self.calculate_time_trace(num + 1, trial)
            output = self.time_traces[num][trial]
        return output

    def update_selected_trials(self, selected_trials):
        """
        Updates the loaded trials so don't have unnecessary loaded trials
        Parameters
        ----------
        selected_trials
            trials to be currently loaded
        """
        new_selected = [x for x, y in enumerate(self.trials_all) if
                        y in selected_trials]
        for trial_num, _ in enumerate(self.trials_all):
            if trial_num in new_selected and trial_num not in self.trials_loaded_time_trace_indices:
                self.dataset_trials_filtered[trial_num] = self.load_trial_filter_step(
                    trial_num).compute()
                self.trials_loaded_time_trace_indices.append(trial_num)
            if trial_num not in new_selected and trial_num in self.trials_loaded_time_trace_indices \
                    and trial_num not in self._trials_loaded_indices:
                self.dataset_trials_filtered[trial_num] = False

            if trial_num not in new_selected and trial_num in self.trials_loaded_time_trace_indices:
                self.trials_loaded_time_trace_indices.remove(trial_num)

    def get_eigen_vector(self, box_num, vector_num, trial_num):
        vector = loadEigenVectors(spatial_box_num=box_num, time_box_num=trial_num,
                                  save_dir=self.save_dir_path).compute()
        spatial_box = SpatialBox(box_num,
                                 total_boxes=self.box_params["total_num_spatial_boxes"],
                                 image_shape=self.shape,
                                 spatial_overlap=self.box_params[
                                                     "spatial_overlap"] // 2)
        return vector[:, vector_num].reshape(spatial_box.shape)

    def genRoiFromPoint(self, point, growth_factor=1.0):
        """
        Generates an roi from an initial point in the image
        Parameters
        ----------
        point : List[int, int]
            2d cordinates of a point in the image
        growth_factor : float
            how much to grow the roi

        Returns
        -------
        the pixels in said roi in there 1d form
        """
        spatial_boxes = [SpatialBox(box_num=x,
                                    total_boxes=self.box_params_processed[
                                        "total_num_spatial_boxes"],
                                    image_shape=self.shape,
                                    spatial_overlap=self.box_params_processed[
                                                        "spatial_overlap"] // 2) for x
                         in
                         range(self.box_params_processed["total_num_spatial_boxes"])]
        boxes_point_in = [x for x in spatial_boxes if x.pointInBox(point)]

        rois = []
        for box in boxes_point_in:
            all_eigen_vectors_list = []
            for temporal_box_num in range(
                    self.box_params_processed["total_num_time_steps"]):
                all_eigen_vectors_list.append(
                    loadEigenVectors(spatial_box_num=box.box_num,
                                     time_box_num=temporal_box_num,
                                     save_dir=self.save_dir_path).compute())

            all_eigen_vectors = np.hstack(all_eigen_vectors_list)
            box_point = box.point_to_box_point(point)
            box_point_1d = box_point[0] * box.shape[1] + box_point[1]
            roi = roi_extract_image(e_vectors=all_eigen_vectors,
                                    original_shape=box.shape, original_2d_vol=None,
                                    merge=False, box_num=box.box_num,
                                    num_rois=1, refinement=True,
                                    num_eigen_vector_select=
                                    self.roi_extraction_params[
                                        "num_eigen_vector_select"],
                                    max_iter=1,
                                    roi_size_min=0,
                                    fill_holes=self.roi_extraction_params[
                                        "fill_holes"],

                                    elbow_threshold_method=True,
                                    elbow_threshold_value=growth_factor,
                                    eigen_threshold_method=
                                    self.roi_extraction_params[
                                        "eigen_threshold_method"],
                                    eigen_threshold_value=
                                    self.roi_extraction_params[
                                        "eigen_threshold_value"],
                                    merge_temporal_coef=
                                    self.roi_extraction_params[
                                        "merge_temporal_coef"],
                                    roi_size_limit=1000,
                                    initial_pixel=box_point_1d,
                                    print_info=False).compute()

            if len(roi) > 0 and box_point_1d in roi[0]:
                rois.append(box.redefine_spatial_cord_1d(roi).compute()[0])
        if len(rois) > 0:
            final_roi = reduce(combine_rois, rois)
        else:
            return []

        return final_roi

    def calculate_statistics(self):
        statistics = [False for x in range(len(self.rois))]
        A = np.zeros([self.edge_roi_image_flat.shape[0], len(self.rois)], dtype=int)
        for num, roi in enumerate(self.rois):
            A[roi, num] = 1
        rval, significant_samples = classify_components_ep(
            self.dataset_trials_filtered_loaded, A,
            np.vstack([self.get_time_trace(x + 1) for x in range(len(self.rois))]))
        pass

    def export(self):
        temp_type = self.time_trace_params["time_trace_type"]
        for time_type in ["Mean", "DeltaF Over F"]:
            for denoise in [True, False]:
                self.time_trace_params["denoise"] = denoise
                self.time_trace_params["time_trace_type"] = time_type
                self.calculate_time_traces()
                if time_type == "Mean" and denoise:
                    self.calculate_statistics()
        self.time_trace_params["time_trace_type"] = temp_type
        self.save_rois(self.rois)
        roi_save_object = []
        spatial_box = SpatialBox(0, 1, image_shape=self.shape, spatial_overlap=0)
        for num, roi in enumerate(self.rois):
            if self.dataset_params["crop_stack"]:

                cords = [[x[0] + self.dataset_params["crop_x"][0],
                          x[1] + self.dataset_params["crop_y"][0]] for x in
                         spatial_box.convert_1d_to_2d(roi)]
            else:
                cords = spatial_box.convert_1d_to_2d(roi)
            curr_roi = {"id": num, "coordinates": cords}
            roi_save_object.append(curr_roi)

        with open(os.path.join(self.save_dir_path, "roi_list.json"), "w") as f:
            json.dump(roi_save_object, f)
        shape = self.edge_roi_image_flat.shape
        roi_image_blob = self.pixel_with_rois_color_flat
        roi_image_outline = np.hstack(
            [self.edge_roi_image_flat,
             np.zeros(shape),
             np.zeros(shape)])
        max_image = np.reshape(
            np.max([self.max_images[x] for x in self._trials_loaded_indices], axis=0),
            (-1, 1))
        roi_image_blob_w_background = roi_image_blob + np.hstack(
            [max_image, max_image, max_image]) / max_image.max() * 255
        roi_image_outline_w_background = roi_image_outline + np.hstack(
            [max_image, max_image, max_image]) / max_image.max() * 255
        self.save_image(roi_image_blob,
                        os.path.join(self.save_dir_path, "roi_blob.png"))
        self.save_image(roi_image_outline,
                        os.path.join(self.save_dir_path, "roi_outline.png"))
        self.save_image(roi_image_outline_w_background,
                        os.path.join(self.save_dir_path, "roi_outline_background.png"))
        self.save_image(roi_image_blob_w_background,
                        os.path.join(self.save_dir_path, "roi_blob_background.png"))
        self.save_image(np.hstack(
            [max_image, max_image, max_image]),
            os.path.join(self.save_dir_path, "max.png"))

    def save_image(self, image, path):
        img = Image.fromarray(
            (np.reshape(image - image.min(),
                        (self.shape[0], self.shape[1], 3)) / (
                     image.max() - image.min()) * 255).astype("uint8"))
        img.save(path)
    def delete_roi_vars(self):
        self.rois = []
        self.rois_loaded = False
        self.roi_max_cord_list = None
        self.roi_min_cord_list = None
        self.pixel_with_rois_flat = None
        self.pixel_with_rois_color_flat = None
        self.time_traces = []
        self.edge_roi_image_flat = None
        self.pixel_with_rois_color = None
        self.eigen_norm_image = None



import sys

from qtpy import QtCore
from qtpy.QtWidgets import *


class Stream(QtCore.QObject):
    newText = QtCore.Signal(str)

    def __init__(self):
        super().__init__()

        # print(type(self.newText))
        # self.connect(self.newText)

    def write(self, text):
        self.newText.emit(str(text))

    def flush(self):
        pass


class ConsoleWidget(QWidget):
    def __init__(self):
        super().__init__()
        self.layout = QHBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.label_start = QLabel()
        self.label_start.setStyleSheet("font-size: 15pt")
        self.label = QLabel()
        self.label.setStyleSheet("font-size: 15pt")
        self.label.setText("")
        self.progress_bar = QProgressBar()
        self.progress_bar.setStyleSheet("font-size: 15pt")
        self.progress_bar.hide()
        self.label_start.hide()
        self.layout.addWidget(self.label_start, alignment=QtCore.Qt.AlignRight,
                              stretch=20)
        self.layout.addWidget(self.progress_bar, alignment=QtCore.Qt.AlignRight,
                              stretch=1)
        self.layout.addWidget(self.label, alignment=QtCore.Qt.AlignRight, stretch=1)

        # self.setMinimumHeight(100)
        # self.process = QTextBrowser()
        # self.process.moveCursor(QtGui.QTextCursor.Start)
        # self.process.ensureCursorVisible()
        # self.process.setLineWrapColumnOrWidth(500)
        # self.process.setLineWrapMode(QTextEdit.FixedPixelWidth)
        #
        # self.layout.addWidget(self.process)
        self.setLayout(self.layout)
        # sys.stdout = Stream()
        # sys.stdout.newText.connect(self.onUpdateText)
        self.update = True

    def updateText(self, text, warning=False):
        # self.label.show()
        self.progress_bar.hide()
        self.label.show()
        self.label_start.hide()
        self.label.setText(text)
        if warning:
            self.label.setStyleSheet("font-size: 15pt;background-color:#CC2936;")
        else:
            self.label.setStyleSheet("font-size: 15pt;")
        # if self.update:
        #     self.update =False
        #     print(text)
        #     self.update = True

    def updateProgressBar(self, start_text, percent):
        self.label_start.show()
        self.label.hide()
        self.progress_bar.show()
        self.label_start.setText(start_text)
        self.progress_bar.setValue(percent)
    def __del__(self):
        sys.stdout = sys.__stdout__



import logging
import sys

from qtpy import QtWidgets

from CIDAN.GUI.Data_Interaction.Signals import StrIntSignal
from CIDAN.GUI.Data_Interaction.Thread import Thread

logger = logging.getLogger("CIDAN.GUI.Data_Interaction.PreprocessThread")


class PreprocessThread(Thread):

    def __init__(self, main_widget, button, preprocess_tab):
        super().__init__(main_widget)
        self.button = button
        self.preprocess_tab = preprocess_tab
        self.reportProgress = StrIntSignal()
        self.reportProgress.sig.connect(main_widget.console.updateProgressBar)
        self.signal.sig.connect(lambda x: self.endThread(x))

    def run(self):
        if self.main_widget.dev:

            self.data_handler.calculate_filters(self.reportProgress)
            self.signal.sig.emit(True)
        else:
            try:
                self.data_handler.calculate_filters(self.reportProgress)
                self.signal.sig.emit(True)
            except Exception as e:
                logger.error(e)
                print("Unexpected error:", sys.exc_info()[0])
                self.main_widget.console.updateText("Unexpected error: " +
                                                    sys.exc_info()[0])
                error_dialog = QtWidgets.QErrorMessage()
                error_dialog.showMessage("Unexpected error: " + str(e))
                self.signal.sig.emit(False)

    def runThread(self):

        if not any([x.isRunning() for x in self.main_widget.thread_list]):
            print("Starting preprocessing sequence")
            self.main_widget.console.updateText("Starting preprocessing sequence")
            # self.button.setEnabled(False)
            self.start()
        else:
            print(
                "Previous process in process, please wait to start new one till "
                "finished")

    def endThread(self, success):
        self.button.setEnabled(True)
        if success:
            print("Finished preprocessing sequence")

            self.main_widget.updateTabs()
            self.main_widget.console.updateText("Finished preprocessing sequence")



import numpy as np

from CIDAN.TimeTrace.mean import calculateMeanTrace


def calculateDeltaFOverF(roi, data, denoise=True, given_mean=False):
    """
    Calcualtes the DeltaF Over F time trace for a given roi and data
    Parameters
    ----------
    roi : np.ndarray or list
        list of pixels for data (1D cords)
    data : np.ndarray
        2D list of data for a single trials with pixel num as dim 0, time as dim 1
    denoise : bool
        Whether to denoise mean floresence of data using wavelets
    Returns
    -------
    A 1d nd array with the time trace for the ROI
    """
    if not given_mean:
        time_trace = calculateMeanTrace(roi, data, denoise)
    else:
        time_trace = data
    bottom_10_number = np.percentile(time_trace, 10)

    bottom_10_frames = time_trace[time_trace < bottom_10_number]
    bottom_10_avg = np.mean(bottom_10_frames)

    if bottom_10_avg == 0:
        return time_trace
    return (time_trace - bottom_10_avg) / bottom_10_avg



import os
from typing import Tuple

import hnswlib as hnsw
import numpy as np
# os.environ["OMP_NUM_THREADS"] = "10"  # export OMP_NUM_THREADS=4
# os.environ["OPENBLAS_NUM_THREADS"] = "10"  # export OPENBLAS_NUM_THREADS=4
# os.environ["MKL_NUM_THREADS"] = "10"  # export MKL_NUM_THREADS=6
# os.environ["VECLIB_MAXIMUM_THREADS"] = "10"  # export VECLIB_MAXIMUM_THREADS=4
# os.environ["NUMEXPR_NUM_THREADS"] = "10"  # export NUMEXPR_NUM_THREADS=6
import scipy.sparse as sparse
from dask import delayed

from CIDAN.LSSC.functions.progress_bar import printProgressBarROI


# from matplotlib import pyplot as plt


@delayed
def calcAffinityMatrix(*, pixel_list: np.ndarray, metric: str, knn: int,
                       accuracy: int, connections: int, normalize_w_k: int,
                       num_threads:
                       int, spatial_box_num: int, temporal_box_num: int,
                       total_num_spatial_boxes: int, total_num_time_steps: int,
                       save_dir: str, progress_signal=None):
    """
    Calculates an pairwise affinity matrix for the image stack
    Parameters
    ----------
    pixel_list: a 2d np array of pixels with dim 0 being a list of pixels and dim 1 being pixel values over time
    metric: can be "l2" squared l2, "ip" Inner product, "cosine" Cosine similarity
    knn: number of nearest neighbors to search for
    accuracy: time of construction vs acuracy tradeoff
    connections: max number of outgoing connections
    num_threads: number of threads to use
    normalize_w_k: kth clostest neighbor for autotune

    Returns
    -------
    affinity matrix
    """
    # TODO make connections value scale based on available memory

    dim = pixel_list.shape[1]
    num_elements = pixel_list.shape[0]
    # print("Spatial Box {}, Time Step {}: Started Processing".format(spatial_box_num,
    #                                                                 temporal_box_num))
    knn_graph = hnsw.Index(space=metric, dim=dim)
    knn_graph.init_index(max_elements=num_elements, ef_construction=accuracy,
                         M=connections)
    knn_graph.add_items(pixel_list, num_threads=num_threads)
    indices, distances = knn_graph.knn_query(pixel_list, k=int(knn),
                                             num_threads=num_threads)
    # import faiss  # make faiss available
    # index = faiss.IndexFlatL2(dim)
    # pixel_list = np.ascontiguousarray(pixel_list.astype('float32'))
    # index.add(pixel_list)  # add vectors to the index
    # distances, indices = index.search(pixel_list, knn)
    # distances = np.power(distances,2)

    # lazy random walk means it returns distance of zero for same point
    # nbrs = LSHForest(n_estimators=20, n_candidates=200,
    #                  n_neighbors=10).fit(pixel_list)
    # distances, indices = nbrs.kneighbors(pixel_list)

    # from annoy import AnnoyIndex
    # t = AnnoyIndex(dim, 'dot')  # Length of item vector that will be indexed
    # for num,x in enumerate(pixel_list):
    #     t.add_item(num,x)
    # t.build(20)
    # indices = np.zeros((num_elements, knn), dtype=int)
    # distances = np.zeros((num_elements, knn))
    # for num in range(num_elements):
    #     indice, distance = t.get_nns_by_item(num, knn, search_k=-1, include_distances=True)
    #     indices[num] = indice
    #     distances[num] = distance
    # distances = np.power(distances,2)
    # indices = io.loadmat('/Users/sschickler/Code_Devel/LSSC-python/inputs/nf_0200_indices.mat')['Inds'].transpose((1, 0))-1
    # distances = io.loadmat('/Users/sschickler/Code_Devel/LSSC-python/inputs/nf_0200_distances.mat')['Dis'].transpose((1, 0))
    # print("Lodaed mat files")
    # print(indices.shape)
    # print(distances.shape)
    # print(num_elements)
    # TODO add comments here
    reformat_indicies_x = np.repeat(np.arange(0, num_elements, 1), knn)
    reformat_indicies_y = np.reshape(indices[:, 0:knn], (-1))
    reformat_distances = np.reshape(distances[:, 0:knn], (-1))

    # Self tuning adaptive bandwidth
    scale_factor_indices = np.repeat(distances[:, normalize_w_k], knn) + .000000001
    scale_factor_2_per_distances = np.power(scale_factor_indices[reformat_indicies_x],
                                            .5) * \
                                   np.power(scale_factor_indices[reformat_indicies_y],
                                            .5)
    reformat_distances_scaled = np.exp(
        -reformat_distances / scale_factor_2_per_distances)
    # TODO change to go direct to csr matrix
    reformat_distances_scaled[reformat_indicies_x == reformat_indicies_y] = 0
    K = sparse.csr_matrix(
        (
            reformat_distances_scaled,
            (reformat_indicies_x,
             reformat_indicies_y)),
        shape=(num_elements, num_elements))
    K_sym = (K + K.transpose()) / 2
    # K_sym[np.arange(0, num_elements, 1),np.arange(0, num_elements, 1)] = 0
    # save_image(K_sym.todense(),0)
    # print(str(temporal_box_num)+" "+str(spatial_box_num) + "  " + str(
    #     np.count_nonzero(K_sym.diagonal()) / K_sym.diagonal().shape[0]))
    with open(os.path.join(save_dir, "temp_files/embedding/s_%s_t_%s" % (
    str(spatial_box_num), str(temporal_box_num))), "w") as f:
        f.write("done")
    printProgressBarROI(total_num_spatial_boxes, total_num_time_steps, save_dir,
                        progress_signal=progress_signal)
    return K_sym


def save_image(image, num):
    pass
    # print(image.diagonal())
    # plt.imshow(
    #     image, aspect="auto")
    # plt.colorbar()
    # path = os.path.join("/Users/sschickler/Code_Devel/LSSC-python/input_images/test",
    #                     "eigen{}.png".format(
    #                         num))
    # plt.savefig(path, dpi=300)
    # plt.close()
    # image
    # img.save(image_path)
def calcDInv(K: sparse.csr_matrix):
    """Calculates a scaling diagonal matrix D to rescale eigen vectors

    Parameters
    ----------
    K the sparse matrix K for the pairwise affinity

    Returns
    -------
    a sparse matrix with type csr, and D's diagonal values
    """
    dim = K.shape[0]
    D_diag_inv = 1 / (
                K.sum(axis=1) + .000000001)  # add small epsilon to each row in K.sum()

    # D_diag = 1 / K.sum(axis=1)
    # print("D_diag",D_diag)
    D_sparse = sparse.dia_matrix((np.reshape(D_diag_inv, [1, -1]), [0]),
                                 (dim, dim))
    return sparse.csr_matrix(D_sparse), (K.sum(axis=1) + .000000001)


def calcLaplacian(P_sparse: sparse.csr_matrix, dim: Tuple[int, int]):
    """
    Calculates the Laplacian
    Parameters
    ----------
    P_sparse
        A sparse matrix of the normalized affinity matrix
    dim
        dimensions of P_sparse
    Returns
    -------
    Laplacian matrix
    # TODO I'm a little unsure what are the dimensions here
    """
    I_sparse = sparse.identity(dim, format="csr")
    laplacian_sparse = I_sparse - P_sparse
    return laplacian_sparse


def calcDSqrt(D_diag: np.ndarray):
    """
    Calculates sparse matrix of the sqrt of D
    Parameters
    ----------
    D_diag
        A list of the diagonal indices of D a the normalization matrix for P

    Returns
    -------
    A sparse CSR matrix of the sqrt of D
    """
    dim = D_diag.shape[0]
    D_sqrt = sparse.csr_matrix(
        sparse.dia_matrix((np.reshape(np.power(D_diag, .5), [1, -1]), [0]),
                          (dim, dim)))
    return D_sqrt


def calcDNegSqrt(D_diag):
    """
    Calculates sparse matrix of the neg sqrt of D
    Parameters
    ----------
    D_diag
        A list of the diagonal indices of D a the normalization matrix for P

    Returns
    -------
    A sparse CSR matrix of the neg sqrt of D
    """
    dim = D_diag.shape[0]
    D_neg_sqrt = sparse.csr_matrix(
        sparse.dia_matrix(
            (np.reshape(np.power(D_diag, -.5), [1, -1]), [0]),
            (dim, dim)))
    return D_neg_sqrt


def embedEigenSqrdNorm(eigen_vectors: np.ndarray) -> np.ndarray:
    """
    Embeds all pixels in image to a set of eigen vectors
    Parameters
    ----------
    eigen_vectors A set of eigen vectors

    Returns
    -------
    A list of pixel values if they where represented by those eigen vectors
    """
    pixel_embedings = np.sum(np.power(eigen_vectors, 2),
                             axis=1)
    return pixel_embedings



from typing import List

from CIDAN.GUI.ListWidgets.ROIListModule import *


class Column(QWidget):  # just used for stylesheets
    pass


class Tab(QWidget):
    def __init__(self, name, column_1: List[QFrame], column_2: List[QFrame],
                 column_2_display=True, column2_moveable=False, horiz_moveable=False):
        super().__init__()
        self.name = name
        self.column_1 = column_1
        self.column_2 = column_2
        self.setMinimumHeight(500)
        if horiz_moveable:
            self.layout = QSplitter(Qt.Horizontal)
            self.layout.setStyleSheet(
                "QSplitter { background-color: #19232D; } ")
        else:
            self.layout = QHBoxLayout()  # Main layout class
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.column_1_widget = Column()

        self.column_1_layout = QVBoxLayout()  # Layout for column 1
        self.column_1_layout.setContentsMargins(0, 0, 0, 0)
        self.column_1_widget.setLayout(self.column_1_layout)
        self.column_1_widget.setStyleSheet("Column { border:1px solid rgb(50, 65, "
                                           "75);} ")
        for module in column_1:
            if issubclass(module.__class__, QWidget):
                self.column_1_layout.addWidget(module)
            else:
                self.column_1_layout.addLayout(module)
        if horiz_moveable:
            self.layout.addWidget(self.column_1_widget)

        else:
            self.layout.addWidget(self.column_1_widget, stretch=1)
        if column_2_display:
            if column2_moveable:
                self.column_2_slider = QSplitter(Qt.Vertical)
                self.column_2_slider.setStyleSheet(
                    "QSplitter { background-color: #19232D; } ")
                for module in column_2:
                    self.column_2_slider.addWidget(module)
                if horiz_moveable:
                    self.layout.addWidget(self.column_2_slider)
                else:
                    self.layout.addWidget(self.column_2_slider, stretch=15)
            else:
                self.column_2_layout = QVBoxLayout()  # Layout for column 2
                self.column_2_layout.setContentsMargins(0, 0, 0, 0)

                self.column_2_widget = Column()
                self.column_2_widget.setStyleSheet(
                    "Column { border:1px solid rgb(50, 65, "
                    "75);} ")
                self.column_2_widget.setLayout(self.column_2_layout)
                for module in column_2:
                    self.column_2_layout.addWidget(module)
                if horiz_moveable:
                    self.layout.addWidget(self.column_2_widget)
                else:
                    self.layout.addWidget(self.column_2_widget, stretch=15)
        if horiz_moveable:
            layout = QVBoxLayout()
            layout.addWidget(self.layout)
            self.setLayout(layout)
        else:
            self.setLayout(self.layout)



import logging

import numpy as np
from qtpy import QtCore
from qtpy.QtCore import *
from qtpy.QtWidgets import *

from CIDAN.GUI.ImageView.ImageViewModule import ImageViewModule
from CIDAN.GUI.Inputs.OptionInput import OptionInput

logger1 = logging.getLogger("CIDAN.ImageView.ROIImageViewModule")


class ROIImageViewModule(ImageViewModule):
    # QApplication.mouseButtons() == Qt.LeftButton
    def __init__(self, main_widget, tab, settings_tab=True):
        super(ROIImageViewModule, self).__init__(main_widget, histogram=False)
        self.tab = tab
        self.resetting_view = False  # Way to prevent infinite loops of reset_view
        self.current_foreground_intensity = 30
        self.click_event = False
        self.outlines = True
        self.trial_selector_input = OptionInput(
            "Time Block:", "", self.set_background,
            val_list=self.data_handler.trials_loaded,
            tool_tip="Select Time block to display",
            display_tool_tip=False, default_index=0,
            show_name=True)
        self.set_background("", "Max Image", update_image=False)
        self.image_item.mouseClickEvent = lambda x: self.roi_view_click(x)
        self.image_item.mouseDragEvent = lambda x: self.roi_view_drag(x)


        shape = main_widget.data_handler.shape
        self.select_image_flat = np.zeros([shape[0] * shape[1], 3])
        self.box_selector_enabled = False
        self.box_selector_cords = [(0, 0), (0, 0)]
        self.current_background_name = "Max Image"

        if settings_tab:
            self.layout.removeWidget(self.image_view)
            self.layout.addWidget(self.createTabLayout())

    def createSettings(self):
        self.display_settings_layout = QVBoxLayout()

        display_settings = QWidget()
        display_settings.setLayout(self.display_settings_layout)
        image_chooser = OptionInput("ROI Display type:", "",
                                    on_change_function=self.set_image,
                                    default_index=0,
                                    tool_tip="Choose background to display",
                                    val_list=["Outlines", "Blob"])

        self.display_settings_layout.addWidget(image_chooser)

        self.background_chooser = OptionInput("Background:", "",
                                              on_change_function=self.set_background,
                                              default_index=2,
                                              tool_tip="Choose background to display",
                                              val_list=["Blank Image", "Mean Image",
                                                        "Max Image",
                                                        # "Temporal Correlation Image",
                                                        "Eigen Norm Image"])

        self.display_settings_layout.addWidget(self.background_chooser)
        self.display_settings_layout.addWidget(self.trial_selector_input)

        background_slider_layout = QHBoxLayout()
        label_0 = QLabel("0")
        label_0.setStyleSheet("QWidget {border: 0px solid #32414B;}")
        background_slider_layout.addWidget(label_0)
        # initializes a slider to control how much to blend background image in when
        # blob is view is enabled
        self.background_slider = QSlider(Qt.Horizontal)
        self.background_slider.setStyleSheet("QWidget {border: 0px solid #32414B;}")
        self.background_slider.setMinimum(0)
        self.background_slider.setMaximum(100)
        self.background_slider.setSingleStep(1)
        self.background_slider.valueChanged.connect(
            self.intensitySliderChanged)
        try:
            self.background_slider.setValue(
                self.current_foreground_intensity)
        except AttributeError:
            pass
        background_slider_layout.addWidget(self.background_slider)
        label_10 = QLabel("10")
        label_10.setStyleSheet("QWidget {border: 0px solid #32414B;}")
        background_slider_layout.addWidget(label_10)
        label_overlay = QLabel("Change background intensity:")
        label_overlay.setStyleSheet("QWidget {border: 0px solid #32414B;}")
        self.display_settings_layout.addWidget(label_overlay)
        self.display_settings_layout.addLayout(background_slider_layout)
        return display_settings
    def createTabLayout(self):
        # ROI view tab section
        roi_view_tabs = QTabWidget()
        roi_view_tabs.setStyleSheet("QTabWidget {font-size: 20px;}")
        # Display settings tab
        display_settings = self.createSettings()

        # ROI image view part
        self.setStyleSheet(
            "margin:0px; border:0px  solid rgb(50, 65, "
            "75); padding: 0px;")
        roi_view_widget = QWidget()
        roi_view_widget_layout = QVBoxLayout()
        roi_view_widget_layout.setContentsMargins(0, 0, 0, 0)
        roi_view_widget_layout.addWidget(self.image_view)
        roi_view_widget.setLayout(roi_view_widget_layout)
        roi_view_tabs.addTab(roi_view_widget, "ROI Display")
        roi_view_tabs.addTab(display_settings, "Display Settings")
        return roi_view_tabs

    def settingsLayout(self):
        return self.display_settings_layout

    def intensitySliderChanged(self):
        self.current_foreground_intensity = (
                float(self.background_slider.value()) / 10)

        self.updateImageDisplay()

    def set_background(self, name, func_name, update_image=True, reset_override=False):
        if (not self.resetting_view or reset_override):
            self.main_widget.console.updateText(
                "Setting current background to: %s" % func_name)
            # Background refers to the image behind the rois
            shape = self.main_widget.data_handler.shape
            if func_name == "Mean Image":
                self.current_background = \
                    self.main_widget.data_handler.mean_images[
                        self.data_handler.trials_loaded.index(
                            self.trial_selector_input.current_state())][:].reshape(
                    [-1, 1])
                self.current_background_name = "Mean Image"
            elif func_name == "Max Image":
                self.current_background = self.main_widget.data_handler.max_images[
                                              self.data_handler.trials_loaded.index(
                                                  self.trial_selector_input.current_state())][
                                          :].reshape(
                    [-1, 1])
                self.current_background_name = "Max Image"
            elif func_name == "Blank Image":
                self.current_background = np.zeros([shape[0] * shape[1], 1])
                self.current_background_name = "Blank Image"
            # if func_name == "Temporal Correlation Image":
            #     self.current_background = self.data_handler.temporal_correlation_image.reshape(
            #         [-1, 1])
            elif func_name == "Eigen Norm Image":
                try:
                    self.current_background = self.data_handler.eigen_norm_image.reshape(
                        [-1, 1])
                    self.current_background_name = "Eigen Norm Image"
                except AttributeError:
                    print("Eigen vectors aren't currently generated or valid")


                    self.current_background = self.main_widget.data_handler.max_images[
                                                  self.data_handler.trials_loaded.index(
                                                      self.trial_selector_input.current_state())][
                                              :].reshape(
                        [-1, 1])
                    if update_image:
                        self.updateImageDisplay()
                    self.main_widget.console.updateText(
                        "Can't display Eigen Norm image, Eigen vectors aren't currently generated or valid")

                    self.current_background_name = "Max Image"
                    return
            else:
                self.current_background_name = "Max Image"
                try:
                    self.trial_selector_input.set_default_val()

                    self.current_background = self.main_widget.data_handler.max_images[
                                              self.data_handler.trials_loaded.index(
                                                  self.trial_selector_input.current_state())][
                                          :].reshape(
                        [-1, 1])
                except AttributeError:
                    pass
            try:
                if self.current_background_name != self.background_chooser.current_state():
                    # pass
                    self.background_chooser.set_val(self.current_background_name)
            except AttributeError:
                pass
            if update_image:
                self.updateImageDisplay()

    def set_image(self, name, func_name, update_image=True):
        if (self.main_widget.checkThreadRunning() and self.data_handler.rois_loaded):

            # Background refers to the image behind the rois
            shape = self.main_widget.data_handler.edge_roi_image_flat.shape
            if func_name == "Outlines":
                self.outlines = True
                self.roi_image_flat = np.hstack([self.data_handler.edge_roi_image_flat,
                                                 np.zeros(shape),
                                                 np.zeros(shape)])
            if func_name == "Blob":
                self.outlines = False
                self.roi_image_flat = self.main_widget.data_handler.pixel_with_rois_color_flat

            if update_image:
                self.updateImageDisplay()

    def updateImageDisplay(self, new=False):

        try:
            # new is to determine whether the zoom should be saved
            # TODO add in update with image paint layer
            shape = self.main_widget.data_handler.shape
            # range_list = self.main_widget.roi_image_view.image_view.view.viewRange()
            background_max = np.percentile(self.current_background, 98)
            background_min = np.percentile(self.current_background, 2)
            background_image_scaled = (self.current_foreground_intensity / 10 * (
                    self.current_background - background_min) * 235 / (
                                           (background_max - background_min) if (
                                                                                        background_max - background_min) != 0 else 1))
            background_image_scaled_3_channel = np.hstack(
                [background_image_scaled, background_image_scaled,
                 background_image_scaled])
            if new and not hasattr(self.main_widget.data_handler,
                                   "edge_roi_image_flat"):
                self.image_item.image = background_image_scaled_3_channel.reshape(
                    (shape[0], shape[1], 3))
                self.image_item.updateImage(autoLevels=False)
                self.image_item.setLevels((0, 255))
            elif new:
                # if self.add_image:
                combined = self.roi_image_flat + background_image_scaled_3_channel + self.select_image_flat

                # else:
                #     combined = background_image_scaled + self.select_image_flat
                #     mask = np.any(self.roi_image_flat != [0, 0, 0], axis=1)
                #     combined[mask] = self.roi_image_flat[mask]
                combined_reshaped = combined.reshape((shape[0], shape[1], 3))
                self.image_item.setLevels((0, 255))

                self.tab.image_view.setImage(combined_reshaped)
            else:
                self.image_item.image = background_image_scaled_3_channel.reshape(
                    (shape[0], shape[1], 3))
                self.image_item.updateImage(autoLevels=False)
                self.image_item.setLevels((0, 255))


                # if self.add_image:
                combined = (self.roi_image_flat + self.select_image_flat).reshape(
                    (shape[0], shape[1], 3))
                self.image_item.image += combined

                self.image_item.updateImage(autoLevels=False)

                # self.main_widget.roi_image_view.image_view.view.setRange(xRange=range_list[0],
                #                                                      yRange=range_list[1])
                # range_list = self.main_widget.roi_image_view.image_view.view.viewRange()
                # print(range_list)

            pass
        except AttributeError as e:
            logger1.error(e)

        except ValueError as e:
            if "shape" in e.args[0]:
                self.reset_view()

    def selectRoi(self, num):
        try:
            color_select = (245, 249, 22)
            color_roi = self.main_widget.data_handler.color_list[
                (num - 1) % len(self.main_widget.data_handler.color_list)]
            self.select_image_flat[
                self.main_widget.data_handler.rois[num - 1]] = color_select
            self.updateImageDisplay()
        except AttributeError:
            pass
        except IndexError:
            self.main_widget.console.updateText(
                "Please regenerate ROIs before trying this operation")

            print("Please regenerate ROIs before trying this operation")
        except ValueError as e:
            if "shape" in e.args[0]:
                self.main_widget.console.updateText("Error please try again")
                print("Error please try again")
                self.reset_view()

    def deselectRoi(self, num):
        try:
            color = self.main_widget.data_handler.color_list[
                (num - 1) % len(self.main_widget.data_handler.color_list)]
            shape_flat = self.data_handler.edge_roi_image_flat.shape
            self.select_image_flat[self.main_widget.data_handler.rois[
                num - 1]] = color if not self.outlines \
                else np.hstack([self.data_handler.edge_roi_image_flat,
                                np.zeros(shape_flat),
                                np.zeros(shape_flat)])[
                self.main_widget.data_handler.rois[num - 1]]
            self.updateImageDisplay()
        except ValueError as e:
            if "shape" in e.args[0]:
                self.main_widget.console.updateText("Error please try again")
                print("Error please try again")
                self.reset_view()

    def zoomRoi(self, num):
        """
        Zooms in to a certain roi
        Parameters
        ----------
        num : int
            roi num starts at 1

        Returns
        -------
        Nothing
        """
        num = num - 1

        max_cord = self.main_widget.data_handler.roi_max_cord_list[num] + 15

        min_cord = self.main_widget.data_handler.roi_min_cord_list[num] - 15

        self.image_view.getView().setYRange(min_cord[1],
                                            max_cord[1])
        self.image_view.getView().setXRange(min_cord[0],
                                            max_cord[0])

    def roi_view_click(self, event):
        if event.button() == QtCore.Qt.RightButton:
            if self.image_item.raiseContextMenu(event):
                event.accept()

        if hasattr(self.main_widget.data_handler,
                   "pixel_with_rois_flat") and self.main_widget.data_handler.pixel_with_rois_flat is not None:
            pos = event.pos()

            y = int(pos.x())
            x = int(pos.y())
            self.click_event = True
            pixel_with_rois_flat = self.main_widget.data_handler.pixel_with_rois_flat
            shape = self.main_widget.data_handler.shape
            roi_num = int(pixel_with_rois_flat[shape[1] * x + y])
            # TODO change to int
            if roi_num != 0:
                event.accept()
                self.tab.roi_list_module.set_current_select(roi_num)

    def roi_view_drag(self, event):
        prev = True
        pos = event.pos()

        y = int(pos.x())  # Because of column order thing in image_view
        x = int(pos.y())
        modifiers = QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.ShiftModifier and not self.box_selector_enabled:
            self.box_selector_enabled = True
            self.box_selector_cords = [(x, y), (x, y)]
            prev = False
        if self.box_selector_enabled:
            event.accept()
            if prev:
                try:

                    if (self.box_selector_cords[0][0] < self.box_selector_cords[1][0]):
                        self.image_item.image[
                        self.box_selector_cords[0][0]:self.box_selector_cords[1][0],
                        self.box_selector_cords[0][1]] -= [255, 255, 255]
                        self.image_item.image[
                        self.box_selector_cords[0][0]:self.box_selector_cords[1][0],
                        self.box_selector_cords[1][1]] -= [255, 255, 255]
                    else:
                        self.image_item.image[
                        self.box_selector_cords[1][0]:self.box_selector_cords[0][0],
                        self.box_selector_cords[0][1]] -= [255, 255, 255]
                        self.image_item.image[
                        self.box_selector_cords[1][0]:self.box_selector_cords[0][0],
                        self.box_selector_cords[1][1]] -= [255, 255, 255]
                    if self.box_selector_cords[0][1] < self.box_selector_cords[1][1]:
                        self.image_item.image[self.box_selector_cords[0][0],
                        self.box_selector_cords[0][1]:self.box_selector_cords[1][1]] -= \
                            [255, 255, 255]
                        self.image_item.image[self.box_selector_cords[1][0],
                        self.box_selector_cords[0][1]:self.box_selector_cords[1][1]] -= \
                            [255, 255, 255]
                    else:
                        self.image_item.image[self.box_selector_cords[0][0],
                        self.box_selector_cords[1][1]:self.box_selector_cords[0][1]] -= \
                            [255, 255, 255]
                        self.image_item.image[self.box_selector_cords[1][0],
                        self.box_selector_cords[1][1]:self.box_selector_cords[0][1]] -= \
                            [255, 255, 255]
                except IndexError:
                    pass
            if QApplication.mouseButtons() != Qt.LeftButton:
                self.box_selector_enabled = False
                shape = self.main_widget.data_handler.shape
                rois_image = np.reshape(
                    self.main_widget.data_handler.pixel_with_rois_flat, shape)
                if (self.box_selector_cords[0][0] < self.box_selector_cords[1][0]):
                    rois_image = rois_image[self.box_selector_cords[0][0]:
                                            self.box_selector_cords[1][0]]
                else:
                    rois_image = rois_image[self.box_selector_cords[1][0]:
                                            self.box_selector_cords[0][0]]
                if self.box_selector_cords[0][1] < self.box_selector_cords[1][1]:
                    rois_image = rois_image[:, self.box_selector_cords[0][1]:
                                               self.box_selector_cords[1][1]]
                else:
                    rois_image = rois_image[:, self.box_selector_cords[1][1]:
                                               self.box_selector_cords[0][1]]
                rois_selected = np.unique(rois_image)[1:]
                self.tab.update_time = False
                for x in rois_selected:
                    self.tab.roi_list_module.roi_item_list[
                        int(x) - 1].check_box.setChecked(True)
                self.tab.update_time = True
                self.tab.deselectRoiTime()
                self.box_selector_cords = [(0, 0), (0, 0)]

            else:
                self.box_selector_cords[1] = (x, y)
                try:
                    if (self.box_selector_cords[0][0] < self.box_selector_cords[1][0]):
                        self.image_item.image[
                        self.box_selector_cords[0][0]:self.box_selector_cords[1][0],
                        self.box_selector_cords[0][1]] += [255, 255, 255]
                        self.image_item.image[
                        self.box_selector_cords[0][0]:self.box_selector_cords[1][0],
                        self.box_selector_cords[1][1]] += [255, 255, 255]
                    else:
                        self.image_item.image[
                        self.box_selector_cords[1][0]:self.box_selector_cords[0][0],
                        self.box_selector_cords[0][1]] += [255, 255, 255]
                        self.image_item.image[
                        self.box_selector_cords[1][0]:self.box_selector_cords[0][0],
                        self.box_selector_cords[1][1]] += [255, 255, 255]
                    if self.box_selector_cords[0][1] < self.box_selector_cords[1][1]:
                        self.image_item.image[self.box_selector_cords[0][0],
                        self.box_selector_cords[0][1]:self.box_selector_cords[1][1]] += \
                            [255, 255, 255]
                        self.image_item.image[self.box_selector_cords[1][0],
                        self.box_selector_cords[0][1]:self.box_selector_cords[1][1]] += \
                            [255, 255, 255]
                    else:
                        self.image_item.image[self.box_selector_cords[0][0],
                        self.box_selector_cords[1][1]:self.box_selector_cords[0][1]] += \
                            [255, 255, 255]
                        self.image_item.image[self.box_selector_cords[1][0],
                        self.box_selector_cords[1][1]:self.box_selector_cords[0][1]] += \
                            [255, 255, 255]

                except IndexError:
                    pass
            self.image_item.updateImage()

    def reset_view(self, updateDisplay=True):
        if not any([x.isRunning() for x in
                    self.main_widget.thread_list]) and not self.resetting_view:
            self.resetting_view = True
            if hasattr(self.main_widget.data_handler,
                       "edge_roi_image_flat") and self.main_widget.data_handler.edge_roi_image_flat is not None:
                shape = self.main_widget.data_handler.edge_roi_image_flat.shape
                self.data_handler.save_rois(self.data_handler.rois)
                self.select_image_flat = np.zeros([shape[0] * shape[1], 3])

                if self.outlines:
                    self.roi_image_flat = np.hstack(
                        [self.data_handler.edge_roi_image_flat,
                         np.zeros(shape),
                         np.zeros(shape)])



                else:
                    self.roi_image_flat = self.main_widget.data_handler.pixel_with_rois_color_flat
            if len(self.trial_selector_input.val_list) != len(
                    self.data_handler.trials_loaded) or any(
                    [x != y for x, y in zip(self.trial_selector_input.val_list,
                                            self.data_handler.trials_loaded)]):
                self.trial_selector_input.set_new_options(
                    self.data_handler.trials_loaded)
                self.trial_selector_input.set_default_val()

            self.set_background("", self.current_background_name, update_image=False,
                                reset_override=True)

            if (updateDisplay):
                self.updateImageDisplay(new=True)
            self.resetting_view = False

    def setImage(self, data):

        # if self.already_loaded == False:
        #     print("changed image")
        #     self.already_loaded = True
        #     self.layout.removeWidget(self.no_image_message)
        #     self.no_image_message.deleteLater()
        #     # self.layout.setAlignment(Qt.AlignLeft)
        #     self.image_view = ImageView()
        #
        #     self.layout.addWidget(self.image_view)
        # bottom_5 = np.percentile(data, 5)
        # top_5 = np.percentile(data, 95)
        # top_10 = np.percentile(data, 90)
        # bottom_2 = np.percentile(data, 2)
        # top_2 = np.percentile(data, 98)
        # data[data>top_10] = top_10
        self.image_view.setImage(data, levelMode='mono', autoRange=True,
                                 autoLevels=True, autoHistogramRange=True)
        # self.top_2 = np.percentile(data, 98)
        # self.bottom_5 = np.percentile(data, 5)
        # self.image_view.setLevels(bottom_5, top_2)
        # self.image_view.setHistogramRange(bottom_2, top_2)



from CIDAN.GUI.Data_Interaction.Signals import *


class Worker(QRunnable):
    def __init__(self, main_widget, parent=None):
        super(Worker, self).__init__()
        self.main_widget = main_widget
        self.signal = BoolSignal()

    @property
    def data_handler(self):
        return self.main_widget.data_handler
    def run(self):
        pass

    def endThread(self, success):
        pass

    def __del__(self):
        self.exiting = True
        self.wait()



import logging

from CIDAN.GUI.Data_Interaction.Thread import Thread
from CIDAN.GUI.Data_Interaction.Worker import Worker, Slot

logger = logging.getLogger("CIDAN.GUI.Data_Interaction.ROIExctractionThread")


class ROIExtractionWorker(Worker):
    def __init__(self, main_widget, button, roi_list_module, roi_tab):
        super().__init__(main_widget)
        self.roi_tab = roi_tab
        self.roi_list_module = roi_list_module
        self.button = button

        self.signal.sig.connect(lambda x: self.endThread(x))
    def isRunning(self):
        return False
    def run(self):

        if self.main_widget.dev:

            self.data_handler.calculate_roi_extraction()
            print("Finished ROI extraction")
            self.signal.sig.emit(True)
        else:
            try:
                print("worker")
                self.data_handler.calculate_roi_extraction()
                print("Finished ROI extraction")
                self.signal.sig.emit(True)
            except AssertionError as e:
                if (type(e) == AssertionError):
                    print(e.args[0])
                else:
                    print("Something weird happened please reload and try again")
                self.main_widget.data_handler.global_params[
                    "need_recalc_eigen_params"] = True
                logger.error(str(e))
                self.signal.sig.emit(False)
    #
    def runThread(self):
        if not any([x.isRunning() for x in self.main_widget.thread_list]):
            self.run()
        else:
            print(
                "Previous process in process, please wait to start new one till finished")

    def endThread(self, success):
        self.button.setEnabled(True)
        if success:
            self.main_widget.updateTabs()



import logging
import sys

from qtpy import QtWidgets

from CIDAN.GUI.Data_Interaction.DataHandler import DataHandler
from CIDAN.GUI.Data_Interaction.Signals import StrIntSignal
from CIDAN.GUI.Data_Interaction.Thread import Thread

logger = logging.getLogger("CIDAN.GUI.Data_Interaction.PreprocessThread")


class OpenDatasetThread(Thread):

    def __init__(self, main_widget):
        super().__init__(main_widget)
        self.reportProgress = StrIntSignal()
        self.reportProgress.sig.connect(main_widget.console.updateProgressBar)
        self.signal.sig.connect(lambda x: self.endThread(x))

    def run(self):
        if self.main_widget.dev:
            self.main_widget.data_handler = DataHandler(data_path=self.data_path,
                                                        trials=self.trials,
                                                        save_dir_path=self.save_dir_path,
                                                        save_dir_already_created=self.save_dir_already_created,
                                                        load_into_mem=self.load_into_mem)
            self.main_widget.data_handler.calculate_filters(self.reportProgress)

            self.signal.sig.emit(True)
        else:
            try:

                self.signal.sig.emit(True)
            except Exception as e:
                logger.error(e)
                print("Unexpected error:", sys.exc_info()[0])
                self.main_widget.console.updateText("Unexpected error: " +
                                                    sys.exc_info()[0])
                error_dialog = QtWidgets.QErrorMessage()
                error_dialog.showMessage("Unexpected error: " + str(e))
                self.signal.sig.emit(False)

    def runThread(self, data_path, trials, save_dir_path, save_dir_already_created,
                  load_into_mem):
        self.data_path = data_path
        self.trials = trials
        self.save_dir_path = save_dir_path
        self.save_dir_already_created = save_dir_already_created
        self.load_into_mem = load_into_mem
        if not any([x.isRunning() for x in self.main_widget.thread_list]):
            print("Opening Dataset")
            self.main_widget.console.updateText("Opening Dataset")
            # self.button.setEnabled(False)
            self.start()
        else:
            print(
                "Previous process in process, please wait to start new one till "
                "finished")

    def endThread(self, success):
        self.main_widget.init_w_data()
